#include "api/df_files.h"
#include "datafox_library_iv.h"
#include <vector>
#include <boost/algorithm/string/split.hpp>

DF_API_EXTERN_DECORATE int DF_API_CALL DFCSetupLaden( int connectionId, int deviceAddress, const char *filePath, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( channelBridge->isValidDeviceAddress( deviceAddress ) ? deviceAddress : 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin(  "%s\t%i\t%i\t%s", __func__, connectionId, deviceAddress,
        (filePath == nullptr) ? "" : filePath);

    // zulaessige Bereiche pruefen
    if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false
        || filePath == nullptr || strlen( filePath ) < 5 || errorNumber == nullptr )
    {
        if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false )
        {
            deviceAddress = 0;
        }

        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber(deviceAddress);
        }

        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_CHANNEL_CLOSED );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        
        channelBridge->logEnd( "\t%i\t(%i, %i)", *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Anderer Blocktransfer besteht bereits?
    if ( channelBridge->getBlockTransferState( deviceAddress ) == controller_segment::bts_active 
        && channelBridge->getBlockTransferType( deviceAddress ) != controller_segment::btt_write_setup_file )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_BLOCK_TRANSFER_ACTIVE );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );

        channelBridge->logEnd( "\t%i\t(%i, %i)", *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Setupdaten uebertragen
    int iRes = channelBridge->writeStudioFile( deviceAddress, filePath, errorNumber );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( deviceAddress, *errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( deviceAddress ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", *errorNumber, iRes, channelBridge->getLastErrorNumber( deviceAddress ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCDownload( int connectionId, int deviceAddress, const char *filePath, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( channelBridge->isValidDeviceAddress( deviceAddress ) ? deviceAddress : 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin(  "%s\t%i\t%i\t%s", __func__, connectionId, deviceAddress,
        (filePath == nullptr) ? "" : filePath);

    // zulaessige Bereiche pruefen
    if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false
        || filePath == nullptr || strlen( filePath ) < 5 || errorNumber == nullptr )
    {
        if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false )
        {
            deviceAddress = 0;
        }

        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        }

        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    int iRes;

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_CHANNEL_CLOSED );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        
        channelBridge->logEnd( "\t%i\t(%i, %i)", *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Anderer Blocktransfer besteht bereits?
    if ( channelBridge->getBlockTransferState( deviceAddress ) == controller_segment::bts_active 
        && channelBridge->getBlockTransferType( deviceAddress ) != controller_segment::btt_read_setup_file )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_BLOCK_TRANSFER_ACTIVE );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );

        channelBridge->logEnd( "\t%i\t(%i, %i)", *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Setupdaten aus Geraet holen
    iRes = channelBridge->readStudioFile( deviceAddress, filePath, errorNumber );

    if ( !iRes && (*errorNumber != IV_ERROR_SUCCESS) )
    {
        channelBridge->setLastErrorNumber( deviceAddress, *errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( deviceAddress ) );

        channelBridge->logEnd( "\t%i\t(%i, %i)", *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Setupdaten in Datei geschrieben
    channelBridge->logEnd( "\t%i\t(%i, %i)", *errorNumber, 1, channelBridge->getLastErrorNumber( deviceAddress ) );

    return 1;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCUpload( int connectionId, int deviceAddress, const char *filePath, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( channelBridge->isValidDeviceAddress( deviceAddress ) ? deviceAddress : 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%i", __func__, connectionId, deviceAddress );

    // zulaessige Bereiche pruefen
    if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false
        || filePath == nullptr || errorNumber == nullptr )
    {
        if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false )
        {
            deviceAddress = 0;
        }

        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        }

        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", (filePath == nullptr) ? "" : filePath, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    int iRes;
    std::string fileName( filePath );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_CHANNEL_CLOSED );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        
        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", fileName.c_str(), *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Endung gibt funktion vor.
    // Nur ".dfl"
    size_t pos = fileName.size() - 4;

    if ( fileName[pos] == '.' &&
        (((fileName[pos + 1] == 'd' || fileName[pos + 1] == 'D') &&
        (fileName[pos + 2] == 'f' || fileName[pos + 2] == 'F') &&
        (fileName[pos + 3] == 'l' || fileName[pos + 3] == 'L'))) )
    {
        // Textdaten uebertragen
        iRes = channelBridge->writeTextData( deviceAddress, fileName );
        if ( iRes == IV_ERROR_SUCCESS )
        { // Kein Fehler
            iRes = 1;
        }
        else if ( iRes > IV_ERROR_SUCCESS )
        { // Fehler bei Durchfuehrung des Befehls
            // Modulfehler auswerten.
            channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            iRes = 0;
        }
        else
        { // Fehler waehrend Kommunikation oder Verarbeitung in DLL
            channelBridge->setLastErrorNumber( deviceAddress, -iRes );
            iRes = 0;
        }

        // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
        if ( iRes == 0 )
        {
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
            channelBridge->resetWithError( channelBridge->getLastErrorNumber( deviceAddress ) );
        }
    } 
    else if ( fileName[pos] == '.' &&
        (((fileName[pos + 1] == 'd' || fileName[pos + 1] == 'D') &&
        (fileName[pos + 2] == 'f' || fileName[pos + 2] == 'F') &&
        (fileName[pos + 3] == 'k' || fileName[pos + 3] == 'K'))) )
    {
        // Touchdaten uebertragen
        iRes = channelBridge->writeTouchKeyData( deviceAddress, fileName );
        if ( iRes == IV_ERROR_SUCCESS )
        { // Kein Fehler
            iRes = 1;
        }
        else if ( iRes > IV_ERROR_SUCCESS )
        { // Fehler bei Durchfuehrung des Befehls
            // Modulfehler auswerten.
            if ( (iRes >> 8) & 0xFF ) 
            {
                if ( ((iRes >> 8) & 0xFF) == 2 ) 
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
                } 
                else 
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
                }
            }
            else
            {
                // Modulfehler auswerten.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
            iRes = 0;
        }
        else
        { // Fehler waehrend Kommunikation oder Verarbeitung in DLL
            channelBridge->setLastErrorNumber( deviceAddress, -iRes );
            iRes = 0;
        }

        // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
        if ( iRes == 0 )
        {
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
            channelBridge->resetWithError( channelBridge->getLastErrorNumber( deviceAddress ) );
        }
    } 
    else if ( fileName[pos] == '.' &&
        (((fileName[pos + 1] == 'd' || fileName[pos + 1] == 'D') &&
        (fileName[pos + 2] == 'f' || fileName[pos + 2] == 'F') &&
        (fileName[pos + 3] == 'c' || fileName[pos + 3] == 'C'))) )
    {
        // Farbdaten uebertragen
        iRes = channelBridge->writeDisplayColors( deviceAddress, fileName );
        if ( iRes == IV_ERROR_SUCCESS )
        { // Kein Fehler
            iRes = 1;
        }
        else if ( iRes > IV_ERROR_SUCCESS )
        { // Fehler bei Durchfuehrung des Befehls
            // Modulfehler auswerten.
            if ( (iRes >> 8) & 0xFF ) 
            {
                if ( ((iRes >> 8) & 0xFF) == 2 ) 
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
                } 
                else 
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
                }
            }
            else
            {
                // Modulfehler auswerten.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
            iRes = 0;
        }
        else
        { // Fehler waehrend Kommunikation oder Verarbeitung in DLL
            channelBridge->setLastErrorNumber( deviceAddress, -iRes );
            iRes = 0;
        }

        // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
        if ( iRes == 0)
        {
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
            channelBridge->resetWithError( channelBridge->getLastErrorNumber( deviceAddress ) );
        }
    } 
    else 
    {
        // Daten uebertragen
        iRes = channelBridge->writeFileDevice( deviceAddress, fileName, errorNumber );

        // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
        if ( iRes == 0 )
        {
            channelBridge->setLastErrorNumber( deviceAddress, *errorNumber );
            channelBridge->resetWithError( channelBridge->getLastErrorNumber( deviceAddress ) );
        }
    }

    channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", fileName.c_str(), *errorNumber, iRes, channelBridge->getLastErrorNumber( deviceAddress ) );

    if ( (iRes == 0) && (channelBridge->getLastErrorNumber( deviceAddress ) == IV_ERROR_DEVICEFILE_MATCH) )
    {
        // Zusatzinfo in Logdatei schreiben
        int count = 0;
        std::string errorText( channelBridge->GetFileDeviceError() );
        std::vector<std::string> result;
        boost::split( result, errorText, [](char c) { return c == '\n'; } );
        for ( std::string &token : result )
        {
            channelBridge->logBegin( "ERROR.%s\t%i\t%i\t%s", __func__, connectionId, ++count, token.data() );
            channelBridge->logEnd( "" );
        }

        // Hardwareinfo als zusaetzliche Information rausgeben
        errorText = channelBridge->GetFileDeviceInfo();
        if ( errorText.size() )
        {
            count = 0;
            result.clear();
            boost::split( result, errorText, [](char c) { return c == '\n'; } );
            for ( std::string &token : result )
            {
                channelBridge->logBegin( "INFO.%s\t%i\t%i\t%s", __func__, connectionId, ++count, token.data() );
                channelBridge->logEnd( "" );
            }
        }
    }

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCUploadModule( int connectionId, int deviceAddress, int fileType, const char *filePath, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( channelBridge->isValidDeviceAddress( deviceAddress ) ? deviceAddress : 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin("%s\t%i\t%i\t%i", __func__, connectionId, deviceAddress, fileType);

    // zulaessige Bereiche pruefen
    if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false
        || fileType != 0 || filePath == nullptr || errorNumber == nullptr )
    {
        if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false )
        {
            deviceAddress = 0;
        }

        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        }

        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", (filePath == nullptr) ? "" : filePath, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    int iRes;
    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_CHANNEL_CLOSED );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        
        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", filePath, *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Dateityp ermitteln
    struct stat st;
    if ( stat( filePath, &st ) == -1 )
    {
        *errorNumber = IV_ERROR_FILE_OPEN;
        return 0;
    }
    // Nur ".up3" && ".up4" && ".up5" gueltig.
    size_t pos = strlen( filePath ) - 4;
    if (
#if defined(_MSC_VER) || defined(__MINGW32__)
        !((st.st_mode & _S_IFMT) == _S_IFREG) 
#else
        !((st.st_mode & S_IFMT) == S_IFREG) 
#endif
        || filePath[pos] != '.' || 
        !(filePath[pos + 1] == 'u' || filePath[pos + 1] == 'U') ||
        !(filePath[pos + 2] == 'p' || filePath[pos + 2] == 'P') ||
        !(filePath[pos + 3] == '3' || filePath[pos + 3] == '4' || filePath[pos + 3] == '5') )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FILE_EXTENSION );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );

        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", filePath, *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Daten uebertragen
    iRes = channelBridge->writeModule( deviceAddress, fileType, filePath, errorNumber );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( deviceAddress, *errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( deviceAddress ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", filePath, *errorNumber, iRes, channelBridge->getLastErrorNumber( deviceAddress ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCModifyStudioFile( const char *sourceFilePath, const char *destFilePath, const char *xmlData, int flags, int *errorNumber )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s", __func__ );

    // zulaessige Bereiche pruefen
    if ( sourceFilePath == nullptr || destFilePath == nullptr || xmlData == nullptr || errorNumber == nullptr )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( 0 );
        }

        channelBridge->logEnd( "\t%s\t%s\t%s\t%i\t(%i, %i)", (sourceFilePath != nullptr) ? sourceFilePath : "", (destFilePath != nullptr) ? destFilePath : "", (xmlData != nullptr) ? xmlData : "", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;
    
    int iRes = channelBridge->modifyStudioFile( sourceFilePath, destFilePath, xmlData, flags, errorNumber );
        // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( 0, *errorNumber );
    }

    channelBridge->logEnd( "\t%s\t%s\t%s\t%i\t(%i, %i)", sourceFilePath, destFilePath, xmlData, *errorNumber, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCFileUpload( int connectionId, int deviceAddress, int fileType, const char *filePath, int flags, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( channelBridge->isValidDeviceAddress( deviceAddress ) ? deviceAddress : 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%i\t%i", __func__, connectionId, deviceAddress, fileType );

    // zulaessige Bereiche pruefen
    if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false 
        || filePath == nullptr || errorNumber == nullptr 
        || fileType < -1 || fileType > 4 )
    {
        if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false )
        {
            deviceAddress = 0;
        }

        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        }

        channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", (filePath == nullptr) ? "" : filePath, flags, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    int iRes;
    char _fileName[0x200];
    memset( _fileName, 0, sizeof( _fileName ) );
    strncpy( _fileName, filePath, sizeof( _fileName ) - 1 );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_CHANNEL_CLOSED );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        
        channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", filePath, flags, *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Blocktransfer besteht bereits? Nur bei Uebertragung von Dateidaten wird extra geprüft.
    if ( channelBridge->getBlockTransferState( deviceAddress ) == controller_segment::bts_active && fileType != 4 )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_BLOCK_TRANSFER_ACTIVE );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );

        channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", filePath, flags, *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Endung gibt funktion vor.
    // Nur ".dfl"
    size_t pos = strlen( _fileName ) - 4;

    if ( fileType == 0 || (fileType == -1 && _fileName[pos] == '.' 
        && (_fileName[pos + 1] == 'd' || _fileName[pos + 1] == 'D') 
        && (_fileName[pos + 2] == 'f' || _fileName[pos + 2] == 'F') 
        && (_fileName[pos + 3] == 'l' || _fileName[pos + 3] == 'L')) )
    {
        // Textdaten uebertragen
        iRes = channelBridge->writeTextData( deviceAddress, _fileName );
        if ( iRes == IV_ERROR_SUCCESS )
        {
            // Kein Fehler
            iRes = 1;
        }
        else if ( iRes > IV_ERROR_SUCCESS )
        {
            // Fehler bei Durchfuehrung des Befehls
            // Modulfehler auswerten.
            channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            iRes = 0;
        }
        else
        {
            // Fehler waehrend Kommunikation oder Verarbeitung in DLL
            channelBridge->setLastErrorNumber( deviceAddress, -iRes );
            iRes = 0;
        }
    }
    else if ( fileType == 1 || (fileType == -1 && _fileName[pos] == '.' 
        && (_fileName[pos + 1] == 'd' || _fileName[pos + 1] == 'D') 
        && (_fileName[pos + 2] == 'f' || _fileName[pos + 2] == 'F') 
        && (_fileName[pos + 3] == 'k' || _fileName[pos + 3] == 'K')) )
    {
        // Touchdaten uebertragen
        iRes = channelBridge->writeTouchKeyData( deviceAddress, _fileName );
        if ( iRes == IV_ERROR_SUCCESS )
        {
            // Kein Fehler
            iRes = 1;
        }
        else if ( iRes > IV_ERROR_SUCCESS )
        {
            // Fehler bei Durchfuehrung des Befehls
            // Modulfehler auswerten.
            if ( (iRes >> 8) & 0xFF )
            {
                if ( ((iRes >> 8) & 0xFF) == 2 ) 
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
                }
                else 
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
                }
            }
            else
            {
                // Modulfehler auswerten.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }

            iRes = 0;
        }
        else
        {
            // Fehler waehrend Kommunikation oder Verarbeitung in DLL
            channelBridge->setLastErrorNumber( deviceAddress, -iRes );
            iRes = 0;
        }
    }
    else if ( fileType == 2 || (fileType == -1 && _fileName[pos] == '.' 
        && (_fileName[pos + 1] == 'd' || _fileName[pos + 1] == 'D') 
        && (_fileName[pos + 2] == 'f' || _fileName[pos + 2] == 'F') 
        && (_fileName[pos + 3] == 'c' || _fileName[pos + 3] == 'C')) )
    {
        // Farbdaten uebertragen
        iRes = channelBridge->writeDisplayColors( deviceAddress, filePath );
        if ( iRes == IV_ERROR_SUCCESS )
        {
            // Kein Fehler
            iRes = 1;
        }
        else if ( iRes > IV_ERROR_SUCCESS )
        {
            // Fehler bei Durchfuehrung des Befehls
            // Modulfehler auswerten.
            if ( (iRes >> 8) & 0xFF )
            {
                if ( ((iRes >> 8) & 0xFF) == 2 )
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
                }
                else 
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
                }
            }
            else
            {
                // Modulfehler auswerten.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }

            iRes = 0;

        }
        else
        {
            // Fehler waehrend Kommunikation oder Verarbeitung in DLL
            channelBridge->setLastErrorNumber( deviceAddress, -iRes );
            iRes = 0;
        }
    }
    else if ( fileType == 3 || (fileType == -1 && _fileName[pos] == '.' 
        && (_fileName[pos + 1] == 'd' || _fileName[pos + 1] == 'D') 
        && (_fileName[pos + 2] == 'f' || _fileName[pos + 2] == 'F') 
        && (_fileName[pos + 3] == '0' || _fileName[pos + 3] == '0')) )
    {
        // WLAN-Konfiguration uebertragen
        iRes = channelBridge->writeWlanConfigData( deviceAddress, _fileName );
        if ( iRes == IV_ERROR_SUCCESS )
        {
            // Kein Fehler
            iRes = 1;
        }
        else if ( iRes > IV_ERROR_SUCCESS )
        {
            // Fehler bei Durchfuehrung des Befehls
            // Modulfehler auswerten.
            if ( (iRes >> 8) & 0xFF )
            {
                if ( ((iRes >> 8) & 0xFF) == 2 )
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
                }
                else 
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
                }
            }
            else
            {
                // Modulfehler auswerten.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }

            iRes = 0;
        }
        else
        {
            // Fehler waehrend Kommunikation oder Verarbeitung in DLL
            channelBridge->setLastErrorNumber( deviceAddress, -iRes );
            iRes = 0;
        }
    }
    else if ( (fileType == -1 && _fileName[pos] == '.' 
        && (_fileName[pos + 1] == 'd' || _fileName[pos + 1] == 'D') 
        && (_fileName[pos + 2] == 'f' || _fileName[pos + 2] == 'F') 
        && (_fileName[pos + 3] == '1' || _fileName[pos + 3] == '1')) )
    {
        // Hardwarekonfiguration uebertragen
        iRes = channelBridge->writeHardwareConfigData( deviceAddress, _fileName );
        if ( iRes == IV_ERROR_SUCCESS )
        {
            // Kein Fehler
            iRes = 1;
        }
        else if ( iRes > IV_ERROR_SUCCESS )
        {
            // Fehler bei Durchfuehrung des Befehls
            // Modulfehler auswerten.
            if ( (iRes >> 8) & 0xFF )
            {
                if ( ((iRes >> 8) & 0xFF) == 2 )
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
                }
                else 
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
                }
            }
            else
            {
                // Modulfehler auswerten.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }

            iRes = 0;
        }
        else
        {
            // Fehler waehrend Kommunikation oder Verarbeitung in DLL
            channelBridge->setLastErrorNumber( deviceAddress, -iRes );
            iRes = 0;
        }
    }
    else if ( fileType == 4 )
    {
        // Anderer Blocktransfer besteht bereits?
        if ( channelBridge->getBlockTransferState( deviceAddress ) == controller_segment::bts_active 
            && channelBridge->getBlockTransferType( deviceAddress ) != controller_segment::btt_write_file )
        {
            channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_BLOCK_TRANSFER_ACTIVE );
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );

            channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", filePath, flags, *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
            return 0;
        }

        // Datei beliebiger Daten uebertragen
        iRes = channelBridge->writeFileData( deviceAddress, _fileName );
        if ( iRes == IV_ERROR_SUCCESS )
        {
            // Kein Fehler
            iRes = 1;
        }
        else if ( iRes > IV_ERROR_SUCCESS )
        {
            // Fehler bei Durchfuehrung des Befehls
            // Modulfehler auswerten.
            if ( (iRes >> 8) & 0xFF )
            {
                if ( ((iRes >> 8) & 0xFF) == 2 )
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
                }
                else 
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
                }
            }
            else
            {
                // Modulfehler auswerten.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }

            iRes = 0;
        }
        else
        {
            // Fehler waehrend Kommunikation oder Verarbeitung in DLL
            channelBridge->setLastErrorNumber( deviceAddress, -iRes );
            iRes = 0;
        }
    }
    else
    {
        // Angegebener Type oder Dateiendung ist unbekannt.
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FILE_EXTENSION );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( deviceAddress ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", filePath, flags, *errorNumber, iRes, channelBridge->getLastErrorNumber( deviceAddress ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCFileDownload( int connectionId, int deviceAddress, int fileType, const char *filePath, int flags, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionId );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( channelBridge->isValidDeviceAddress( deviceAddress ) ? deviceAddress : 0, IV_ERROR_SUCCESS );

    channelBridge->logBegin( "%s\t%i\t%i\t%i", __func__, connectionId, deviceAddress, fileType );

    // zulaessige Bereiche pruefen
    if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false 
        || filePath == nullptr || errorNumber == nullptr 
        || fileType < -1  || fileType > 4 )
    {
        if ( channelBridge->isValidDeviceAddress( deviceAddress ) == false )
        {
            deviceAddress = 0;
        }

        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_PARAMETER );

        if ( errorNumber )
        {
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        }

        channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", (filePath == nullptr) ? "" : filePath, flags, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Fehlerobjekt initalisieren.
    *errorNumber = IV_ERROR_SUCCESS;

    int iRes;
    char _fileName[0x200];
    memset( _fileName, 0, sizeof( _fileName ) );
    strncpy( _fileName, filePath, sizeof( _fileName ) - 1 );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_CHANNEL_CLOSED );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        
        channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", filePath, flags, *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Blocktransfer besteht bereits?
    if ( channelBridge->getBlockTransferState( deviceAddress ) == controller_segment::bts_active && fileType != 4 )
    {
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_BLOCK_TRANSFER_ACTIVE );
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );

        channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", filePath, flags, *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
        return 0;
    }

    // Endung kann funktion vorgeben.
    size_t pos = strlen( _fileName ) - 4;

    if ( fileType == 0 || (fileType == -1 && _fileName[pos] == '.' 
        && (_fileName[pos + 1] == 'd' || _fileName[pos + 1] == 'D') 
        && (_fileName[pos + 2] == 'f' || _fileName[pos + 2] == 'F') 
        && (_fileName[pos + 3] == 'l' || _fileName[pos + 3] == 'L')) )
    {
        // Textdaten lesen
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
        iRes = 0;
    }
    else if ( fileType == 1 || (fileType == -1 && _fileName[pos] == '.' 
        && (_fileName[pos + 1] == 'd' || _fileName[pos + 1] == 'D') 
        && (_fileName[pos + 2] == 'f' || _fileName[pos + 2] == 'F') 
        && (_fileName[pos + 3] == 'k' || _fileName[pos + 3] == 'K')) )
    {
        // Touchdaten uebertragen
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
        iRes = 0;
    }
    else if ( fileType == 2 || (fileType == -1 && _fileName[pos] == '.' 
        && (_fileName[pos + 1] == 'd' || _fileName[pos + 1] == 'D') 
        && (_fileName[pos + 2] == 'f' || _fileName[pos + 2] == 'F') 
        && (_fileName[pos + 3] == 'c' || _fileName[pos + 3] == 'C')) )
    {
        // Farbdaten uebertragen
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
        iRes = 0;
    }
    else if ( fileType == 3 || (fileType == -1 && _fileName[pos] == '.' 
        && (_fileName[pos + 1] == 'd' || _fileName[pos + 1] == 'D') 
        && (_fileName[pos + 2] == 'f' || _fileName[pos + 2] == 'F') 
        && (_fileName[pos + 3] == '0' || _fileName[pos + 3] == '0')) )
    {
        // WLAN-Konfiguration uebertragen
        iRes = channelBridge->readWlanConfigData( deviceAddress, _fileName );
        if ( iRes == IV_ERROR_SUCCESS )
        {
            // Kein Fehler
            iRes = 1;
        }
        else if ( iRes > IV_ERROR_SUCCESS )
        {
            // Fehler bei Durchfuehrung des Befehls
            // Modulfehler auswerten.
            if ( (iRes >> 8) & 0xFF )
            {
                if ( ((iRes >> 8) & 0xFF) == 2 )
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
                }
                else 
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
                }
            }
            else
            {
                // Modulfehler auswerten.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
            
            iRes = 0;
        }
        else
        {
            // Fehler waehrend Kommunikation oder Verarbeitung in DLL
            channelBridge->setLastErrorNumber( deviceAddress, -iRes );
            iRes = 0;
        }
    }
    else if ( (fileType == -1 && _fileName[pos] == '.' 
        && (_fileName[pos + 1] == 'd' || _fileName[pos + 1] == 'D') 
        && (_fileName[pos + 2] == 'f' || _fileName[pos + 2] == 'F') 
        && (_fileName[pos + 3] == '1' || _fileName[pos + 3] == '1')) )
    {
        // Hardware-Konfiguration Daten uebertragen
        iRes = channelBridge->readHardwareConfigData( deviceAddress, _fileName );
        if ( iRes == IV_ERROR_SUCCESS )
        {
            // Kein Fehler
            iRes = 1;
        }
        else if ( iRes > IV_ERROR_SUCCESS )
        {
            // Fehler bei Durchfuehrung des Befehls
            // Modulfehler auswerten.
            if ( (iRes >> 8) & 0xFF )
            {
                if ( ((iRes >> 8) & 0xFF) == 2 )
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
                }
                else 
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
                }
            }
            else
            {
                // Modulfehler auswerten.
                channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
            }
            
            iRes = 0;
        }
        else
        {
            // Fehler waehrend Kommunikation oder Verarbeitung in DLL
            channelBridge->setLastErrorNumber( deviceAddress, -iRes );
            iRes = 0;
        }
    }
    else if ( fileType == 4 )
    {
        // Anderer Blocktransfer besteht bereits?
        if ( channelBridge->getBlockTransferState( deviceAddress ) == controller_segment::bts_active 
            && channelBridge->getBlockTransferType( deviceAddress ) != controller_segment::btt_read_file )
        {
            channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_BLOCK_TRANSFER_ACTIVE );
            *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );

            channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", filePath, flags, *errorNumber, 0, channelBridge->getLastErrorNumber( deviceAddress ) );
            return 0;
        }

        // Datei mit beliebigen Daten uebertragen
        iRes = channelBridge->readFileData( deviceAddress, _fileName );
        if ( iRes == IV_ERROR_SUCCESS )
        {
            // Kein Fehler
            iRes = 1;
        }
        else if ( iRes > IV_ERROR_SUCCESS )
        {
            // Fehler bei Durchfuehrung des Befehls
            // Modulfehler auswerten.
            if ( (iRes >> 8) & 0xFF )
            {
                if ( ((iRes >> 8) & 0xFF) == 2 )
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FUNCTION_UNAVAILABLE );
                }
                else 
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
                }
            }
            else
            {
                // Modulfehler auswerten.
                if ( iRes == 12 )
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FILE_OPEN );
                }
                else 
                {
                    channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_PROCESSING );
                }
            }

            iRes = 0;
        }
        else
        {
            // Fehler waehrend Kommunikation oder Verarbeitung in DLL
            channelBridge->setLastErrorNumber( deviceAddress, -iRes );
            iRes = 0;
        }
    }
    else
    {
        // Angegebener Type oder Dateiendung ist unbekannt.
        channelBridge->setLastErrorNumber( deviceAddress, IV_ERROR_FILE_EXTENSION );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        *errorNumber = channelBridge->getLastErrorNumber( deviceAddress );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( deviceAddress ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", filePath, flags, *errorNumber, iRes, channelBridge->getLastErrorNumber( deviceAddress ) );

    return iRes;
}
