/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "channel/channel_4_buffer.h"

#ifndef DATAFOX_STUDIO_IV
#include "DatafoxLibraryIV/include/datafox_library_iv.h"
#endif

#include "DatafoxLibraryIV/include/library_version.h"
#include "common/error_defines.h"
#if !defined(_MSC_VER) && !defined(__MINGW32__)
#include <ctype.h>
#endif

#include <stack>
#include <algorithm>
#include <vector>

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Channel4Buffer::Channel4Buffer()
{
    int i;

    // Member Initialisieren
    StartDatensatzB = 0;
    EndDatensatzB = 0;
    MaxDatB = 0;
    MaxListB = 0;

    // Zeiger auf Listenspeicher initialisieren
    for ( i = 0; i < MAX_DATENSATZ_BESCHREIBUNG; ++i )
    {
        m_ucListRecordSize[i] = 0;
        m_listSize[i] = 0;
        m_listSortCheck[i] = false;
        m_pucListMem[i] = NULL;
    }

    // Datensatzgroesse fuer Zutrittslisten (Version 2) initialisieren.
    m_ucEntrance2RecordSize[0] = ENTRANCE2_RECORD_SIZE_0;
    m_ucEntrance2RecordSize[1] = ENTRANCE2_RECORD_SIZE_1;
    m_ucEntrance2RecordSize[2] = ENTRANCE2_RECORD_SIZE_2;
    m_ucEntrance2RecordSize[3] = ENTRANCE2_RECORD_SIZE_3;
    m_ucEntrance2RecordSize[4] = ENTRANCE2_RECORD_SIZE_4;
    m_ucEntrance2RecordSize[5] = ENTRANCE2_RECORD_SIZE_5;
    m_ucEntrance2RecordSize[6] = ENTRANCE2_RECORD_SIZE_6;
    m_ucEntrance2RecordSize[7] = ENTRANCE2_RECORD_SIZE_7;

    // Zeiger auf Zutrittslistenspeicher (Version 2) initialisieren
    for ( i = 0; i < 8; ++i )
    {
        m_ulEntrance2ListSize[i] = 0;
        m_entrance2SortCheck[i] = false;
        m_pucEntrance2ListMem[i] = NULL;
    }

    // Zutrittskontrollstrukturen (Version 2) aufsetzen.
    Entrance2Default();

    // Zeiger auf Timeboylistenspeicher initialisieren
    memset( m_pucTimeboyListMem, 0, sizeof(m_pucTimeboyListMem) );
    ClearTimeboyListBuffer();

    // Speicher des Dokuments initialisieren
    m_file.SetDefault();
    // Aktuelle Library-Version setzen.
    m_file.SetLibraryVersion( libraryVersion );
}

Channel4Buffer::~Channel4Buffer()
{
    reset();
}

//////////////////////////////////////////////////////////////////////
// Member
//////////////////////////////////////////////////////////////////////

void Channel4Buffer::reset()
{
    // Member Initialisieren
    StartDatensatzB = 0;
    EndDatensatzB = 0;
    MaxDatB = 0;
    MaxListB = 0;

    // Reservierten Listenspeicher freigeben
    ClearListBuffer();

    // Reservierten Zutrittslistenspeicher (Version 2) freigeben
    ClearEntrance2ListBuffer();

    // Reservierten Timeboylistenspeicher freigeben
    ClearTimeboyListBuffer();

    // Setupdatei bereinigen
    m_file.SetDefault();
}

void Channel4Buffer::ChangeOrderLong( unsigned char *Buf )
{
    unsigned char h;

    h = Buf[0];
    Buf[0] = Buf[3];
    Buf[3] = h;

    h = Buf[1];
    Buf[1] = Buf[2];
    Buf[2] = h;
}

void Channel4Buffer::ULongToBinMSBFirst( unsigned char *Buf,unsigned int Val )
{
    Buf[3] = (unsigned char)(Val&0xFF);
    Val >>= 8;
    Buf[2] = (unsigned char)(Val&0xFF);
    Val >>= 8;
    Buf[1] = (unsigned char)(Val&0xFF);
    Val >>= 8;
    Buf[0] = (unsigned char)(Val&0xFF);
}

void Channel4Buffer::ChangeOrderInt( unsigned char *Buf )
{
    unsigned char h;
    h = Buf[0];
    Buf[0] = Buf[1];
    Buf[1] = h;
}

void Channel4Buffer::ClearListBuffer()
{
    int i;
    for ( i = 0; i < MAX_DATENSATZ_BESCHREIBUNG; ++i )
    {
        m_listSize[i] = 0;
        if ( m_pucListMem[i] )
        {
            delete[] m_pucListMem[i];
            m_pucListMem[i] = NULL;
        }
    }
}

int Channel4Buffer::MakeListe( int ListNum, int ListCnt, int ListSize, const unsigned char *Buf )
{
    int i;

    // Zuvor reservierten Speicher freigeben
    if ( m_pucListMem[ListNum] )
    {
        delete[] m_pucListMem[ListNum];
        m_pucListMem[ListNum] = NULL;
    }

    // Speicher reservieren
    m_pucListMem[ListNum] = new unsigned char[ListSize + sizeof(unsigned int)];

    if ( m_pucListMem[ListNum] == NULL )
        return 0;

    // Pointer in Tabelle der Listen-Pointer eintragen
    //	*(unsigned int*)&m_pucListMem[ListNum * sizeof(unsigned int)] = m_ulListSize + m_ulListStart;
    //	ChangeOrderLong((unsigned char*)&m_pucListMem[ListNum * sizeof(unsigned int)]);

    // Start fuer Listendaten berechnen
    //	pDat = m_ulListSize + sizeof(unsigned int);	// Platz fuer Anzahl Zeilen freihalten

    // Datenueberlauf auffangen
    /*	if ((pDat + ListSize) > sizeof(m_pucListMem)) {
    // Liste loeschen
    *(unsigned int*)&m_pucListMem[ListNum * sizeof(unsigned int)] = 0;
    return 0;
    }
    */	
    // Listendaten kopieren
    for ( i = 0; i < ListSize; ++i )
        m_pucListMem[ListNum][sizeof(unsigned int) + i] = Buf[i];

    // Anzahl Listenzeilen eintragen
    *(unsigned int*)&m_pucListMem[ListNum][0] = ListCnt;
    //ChangeOrderLong(&m_pucListMem[ListNum][0]);

    // Listen-Pointer auf erstes freies Byte im Listenspeicher setzen
    //m_ulListSize = pDat + ListSize;

    // Liste als unsortiert markieren
    m_listSortCheck[ListNum] = false;

    // Listengrosse fuer weitere Verarbeitung halten
    m_listSize[ListNum] = sizeof(unsigned int) + ListSize;

    return 1;
}

bool Channel4Buffer::comparing( char *left, char *right )
{
    return (strcmp( left, right) < 0);
}

void Channel4Buffer::SortList( unsigned char *pucMem, unsigned char ucRecordSize, unsigned char ucOffs )
{
    std::vector<char*> valPtr;
    // Vector mit Zeigern auf die zu Vergleichenden Werte fuellen.
    unsigned int count = *(unsigned int*)pucMem, i;
    valPtr.resize( count );
    for ( i = 0; i < count; ++i )
        valPtr[i] = (char*)&pucMem[sizeof(unsigned int) + ucOffs + (i * ucRecordSize)];

    // Sortierung durchfuehren.
    stable_sort( valPtr.begin(), valPtr.end(), comparing );

    unsigned char *mem = new unsigned char[sizeof(unsigned int) + count * ucRecordSize];

    *(unsigned int*)mem = count;
    for ( i = 0; i < count; ++i )
        memcpy( &mem[sizeof(unsigned int) + i * ucRecordSize], valPtr.at(i) - ucOffs, ucRecordSize );

    memcpy( pucMem, mem, sizeof(unsigned int) + count * ucRecordSize );
    delete[] mem;
}

void Channel4Buffer::QuicksortList( unsigned char *pucMem, unsigned char ucRecordSize, unsigned char ucOffs, int lLeftRecordNum, int lRightRecordNum )
{
    int lHelp1;
    std::stack<int> sf;

    for ( ; ; )
    {
        while ( lRightRecordNum > lLeftRecordNum )
        {
            lHelp1 = QuicksortListPartition( pucMem, ucRecordSize, ucOffs, lLeftRecordNum, lRightRecordNum );
            if ( (lHelp1 - lLeftRecordNum) > (lRightRecordNum - lHelp1) )
            {
                sf.push( lLeftRecordNum );
                sf.push( lHelp1 - 1 );
                lLeftRecordNum = lHelp1 + 1;
            }
            else
            {
                sf.push( lHelp1 + 1 );
                sf.push( lRightRecordNum );
                lRightRecordNum = lHelp1 - 1;
            }
        }

        if ( sf.empty() )
            break;

        lRightRecordNum = sf.top();
        sf.pop();

        lLeftRecordNum = sf.top();
        sf.pop();
    }
}

int Channel4Buffer::QuicksortListPartition( unsigned char *pucMem, unsigned char ucRecordSize, unsigned char ucOffs, int lLeftRecordNum, int lRightRecordNum )
{
    int lHelp1, lHelp2;

    lHelp1 = lLeftRecordNum - 1;
    lHelp2 = lRightRecordNum;

    for ( ; ; )
    {
        while ( strcmp( (const char*)&pucMem[++lHelp1 * ucRecordSize + sizeof(unsigned int) + ucOffs], 
            (const char*)&pucMem[lRightRecordNum * ucRecordSize + sizeof(unsigned int) + ucOffs]) < 0 );

        while ( strcmp( (const char*)&pucMem[--lHelp2 * ucRecordSize + sizeof(unsigned int) + ucOffs], 
            (const char*)&pucMem[lRightRecordNum * ucRecordSize + sizeof(unsigned int) + ucOffs]) > 0 );

        if ( lHelp1 >= lHelp2 )
            break;

        if ( strcmp( (const char*)&pucMem[lHelp1 * ucRecordSize + sizeof(unsigned int) + ucOffs], 
            (const char*)&pucMem[lHelp2 * ucRecordSize + sizeof(unsigned int) + ucOffs] ) != 0 )
            QuicksortListSwap(pucMem, ucRecordSize, lHelp1, lHelp2);
    }

    if ( strcmp( (const char*)&pucMem[lHelp1 * ucRecordSize + sizeof(unsigned int) + ucOffs], 
        (const char*)&pucMem[lRightRecordNum * ucRecordSize + sizeof(unsigned int) + ucOffs] ) != 0 )
        QuicksortListSwap(pucMem, ucRecordSize, lHelp1, lRightRecordNum);

    return lHelp1;
}

void Channel4Buffer::QuicksortListSwap( unsigned char *pucMem, unsigned char ucRecordSize, int lLeftRecordNum, int lRightRecordNum )
{
    int i;
    unsigned char ucTmp;

    for ( i = 0; i < ucRecordSize; ++i )
    {
        ucTmp = pucMem[lLeftRecordNum * ucRecordSize + sizeof(unsigned int) + i];
        pucMem[lLeftRecordNum * ucRecordSize + sizeof(unsigned int) + i] = pucMem[lRightRecordNum * ucRecordSize + sizeof(unsigned int) + i];
        pucMem[lRightRecordNum * ucRecordSize + sizeof(unsigned int) + i] = ucTmp;
    }
}

void Channel4Buffer::ClearEntrance2ListBuffer( int iListNum )
{
    int i;
    if ( iListNum == -1 )
    {
        // Alle loeschen
        for ( i = 0; i < 8; ++i )
        {
            m_ulEntrance2ListSize[i] = 0;

            // Reservierten Speicher freigeben
            if ( m_pucEntrance2ListMem[i] )
            {
                delete[] m_pucEntrance2ListMem[i];
                m_pucEntrance2ListMem[i] = NULL;
            }
        }
    }
    else
    {
        if ( iListNum > 6 )
            return;

        m_ulEntrance2ListSize[iListNum] = 0;

        // Reservierten Speicher freigeben
        if ( m_pucEntrance2ListMem[iListNum] )
        {
            delete[] m_pucEntrance2ListMem[iListNum];
            m_pucEntrance2ListMem[iListNum] = NULL;
        }
    }
}

int Channel4Buffer::MakeEntrance2List( int ListNum, int ListCnt, int ListSize, const unsigned char *Buf )
{
    int iError = IV_ERROR_SUCCESS;

    // Liste in DLL - Verarbeitungsbuffer schreiben.
    if ( (iError = CheckEntrance2List( ListNum, ListCnt, Buf, ListSize )) != IV_ERROR_SUCCESS )
    {
        if ( iError < 0 )
        {
            // Kommunikationsfehler
            iError = -iError;
        }
        else if ( iError > 0 )
        {
            // Funktionsfehler
            iError = IV_ERROR_PROCESSING;
        }

        return 0;
    }

    // Liste in DLL - Verarbeitungsbuffer schreiben.
    if ( (iError = ImportEntrance2List( ListNum, ListCnt, Buf, ListSize )) != IV_ERROR_SUCCESS )
    {
        if ( iError < 0 )
        {
            // Kommunikationsfehler
            iError = -iError;
        }
        else if ( iError > 0 )
        {
            // Funktionsfehler
            iError = IV_ERROR_PROCESSING;
        }

        return 0;
    }

    // Neu importierte Listen als unsortiert markieren
    m_entrance2SortCheck[ListNum] = false;

    return 1;
}

int Channel4Buffer::CheckEntrance2Dec( const char *buf, int min, int max )
{
    int i; // Alles Ziffern?
    for ( i = 0; i < (int)strlen(buf); ++i )
        if ( !isdigit( buf[i] ) )
            return 0;

    // Bereich eingehalten?
    if ( atoi( buf ) >= min && atoi( buf ) <= max )
        return 1;

    return 0;
}

int Channel4Buffer::CheckEntrance2TM( const char *buf )
{
    char tm[4], stich;
    // Format OK?
    if ( strlen( buf ) != 3 )
        return 0;

    strncpy( tm, buf, 4 );
    stich = atoi( &buf[2] );
    if ( stich > 1 )
        return 0;

    // Bereich eingehalten?
    if ( atoi( tm ) >= 0 && atoi( tm ) <= 31 )
        return 1;
    else if ( atoi( tm ) == 32 && !stich )
        return 1;

    return 0;
}

int Channel4Buffer::CheckEntrance2Date( const char *buf )
{
    // Format OK?
    char date[11];
    int month, day;

    if ( strlen( buf ) != 10 )
        return 0;

    if ( buf[4] != '-' || buf[7] != '-' )
        return 0;

    strncpy( date, buf, 11 );
    date[4] = 0;
    date[7] = 0;

    //year = atoi(date);
    month = atoi( &date[5] );
    day = atoi( &date[8] );
    if ( month < 1 || month > 12 )
        return 0;

    if ( day < 1 || day > 31 )
        return 0;

    return 1;
}

int Channel4Buffer::CheckEntrance2Time( const char *buf )
{
    // Format OK?
    char time[6];
    int hour, minute;

    if ( strlen( buf ) != 5 )
        return 0;

    if ( buf[2] != ':' )
        return 0;

    strncpy( time, buf, 6 );
    time[2] = 0;
    hour = atoi( time );
    minute = atoi( &time[3] );

    if ( hour < 0 || hour > 23 )
        return 0;
    
    if ( minute < 0 || minute > 59 )
        return 0;

    return 1;
}

int Channel4Buffer::CheckEntrance2Weekday( const char *buf )
{
    // Format OK?
    unsigned char wd = 0;
    int i, val;
    if ( strlen( buf ) < 1 || strlen( buf ) > 7 )
        return 0;

    for ( i = 0; i < (int)strlen( buf ); ++i )
    {
        val = buf[i] - 0x30;
        if ( val < 1 || val > 7 )
            return 0;

        if ( wd & (1 << val) )
            return 0;

        wd |= (1 << val);
    }

    return 1;
}

int Channel4Buffer::CheckEntrance2List( unsigned char num, unsigned short recordCount, const unsigned char *buf, unsigned int size )
{
    // ListNum ist:
    // 0: Reader         (33 Byte)
    // 1: Identification (64 Byte)
    // 2: Location       (20 Byte)
    // 3: Time           (25 Byte)
    // 4: Holiday        (21 Byte)
    // 5: Events         (17 Byte)
    // 6: Action         (21 Byte)
    // 7: Action2        (39 Byte)
    int checkRecords = 0;
    unsigned short recordNum;
    unsigned char fieldNum, recordPos = 0;

    // Bereich der gueltigen Nummer pruefen.
    switch ( num )
    {
    case 0:
        // Anzahl moeglicher Datensaetze
        if ( (recordCount < 1) /*|| (recordCount > (FLASH_SIZE_LIST / ENTRANCE_EXT_RECORD_SIZE_0))*/ )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Berechnete Listengroesze pruefen
        if ( size != (unsigned int)(recordCount * ENTRANCE2_RECORD_SIZE_0) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Ueberpruefen der Listendaten durchfuehren?
        if ( checkRecords )
        {
            // Datensaetze pruefen
            for ( recordNum = 0; recordNum < recordCount; ++recordNum )
            {
                // Spalten durchlaufen
                for ( fieldNum = 0; fieldNum < m_entrance2Table[num].count; ++fieldNum )
                {
                    switch ( fieldNum )
                    {
                    case 0: // ID
                    case 1: // ZM
                    case 3: // RefLocation
                    case 4: // RefAction
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos], 1, 9999 ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;
                    case 2: // TM
                        if ( !CheckEntrance2TM( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos] ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;
                    case 5: // PinGeneral
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos], 0, 99999999 ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                    }

                    recordPos += m_entrance2TableField[num][fieldNum].size;
                }
            }
        }
        break;

    case 1:
        if ( (recordCount < 1) /*|| (recordCount > (FLASH_SIZE_LIST / ENTRANCE_EXT_RECORD_SIZE_1))*/ )
            return -IV_ERROR_FUNCTION_PARAMETER;

        if ( size != (unsigned int)(recordCount * ENTRANCE2_RECORD_SIZE_1) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Ueberpruefen der Listendaten durchfuehren?
        if ( checkRecords )
        {
            // Datensaetze pruefen
            for ( recordNum = 0; recordNum < recordCount; ++recordNum )
            {
                // Spalten durchlaufen
                for ( fieldNum = 0; fieldNum < m_entrance2Table[num].count; ++fieldNum )
                {
                    switch ( fieldNum )
                    {
                    case 0: // ID
                        break;

                    case 1: // Group
                    case 3: // Menace
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos], 1, 9999 ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;

                    case 2: // Pin
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos], 0, 99999999 ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;

                    case 4: // ActiveStart
                    case 5: // ActiveEnd
                        if ( !CheckEntrance2Date( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos] ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;

                    case 6: // ActiveGeneral
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos], 0, 1 ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                    }

                    recordPos += m_entrance2TableField[num][fieldNum].size;
                }
            }
        }
        break;

    case 2:
        if ( (recordCount < 1) /*|| (recordCount > (FLASH_SIZE_LIST / ENTRANCE_EXT_RECORD_SIZE_2))*/ )
            return -IV_ERROR_FUNCTION_PARAMETER;

        if ( size != (unsigned int)(recordCount * ENTRANCE2_RECORD_SIZE_2) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Ueberpruefen der Listendaten durchfuehren?
        if ( checkRecords )
        {
            // Datensaetze pruefen
            for ( recordNum = 0; recordNum < recordCount; ++recordNum )
            {
                // Spalten durchlaufen
                for ( fieldNum = 0; fieldNum < m_entrance2Table[num].count; ++fieldNum )
                {
                    switch( fieldNum )
                    {
                    case 0: // ID
                    case 1: // RefGroup
                    case 2: // RefTime
                    case 3: // RefTimeNoPin
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos], 1, 9999 ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;
                    }

                    recordPos += m_entrance2TableField[num][fieldNum].size;
                }
            }
        }
        break;

    case 3:
        if ( (recordCount < 1) /*|| (recordCount > (FLASH_SIZE_LIST / ENTRANCE_EXT_RECORD_SIZE_3))*/ )
            return -IV_ERROR_FUNCTION_PARAMETER;

        if ( size != (unsigned int)(recordCount * ENTRANCE2_RECORD_SIZE_3) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Ueberpruefen der Listendaten durchfuehren?
        if ( checkRecords )
        {
            // Datensaetze pruefen
            for ( recordNum = 0; recordNum < recordCount; ++recordNum )
            {
                // Spalten durchlaufen
                for ( fieldNum = 0; fieldNum < m_entrance2Table[num].count; ++fieldNum )
                {
                    switch ( fieldNum )
                    {
                    case 0: // ID
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos], 1, 9999 ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;

                    case 1: // Weekday
                        if ( !CheckEntrance2Weekday( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos] ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;

                    case 2: // TimeStart
                    case 3: // TimeEnd
                        if  (!CheckEntrance2Time( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos] ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;
                    }

                    recordPos += m_entrance2TableField[num][fieldNum].size;
                }
            }
        }
        break;

    case 4:
        // Anzahl moeglicher Datensaetze
        if ( (recordCount < 1) /*|| (recordCount > (FLASH_SIZE_LIST / ENTRANCE_EXT_RECORD_SIZE_4))*/ )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Berechnete Listengroesze pruefen
        if ( size != (unsigned int)(recordCount * ENTRANCE2_RECORD_SIZE_4) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Ueberpruefen der Listendaten durchfuehren?
        if ( checkRecords )
        {
            // Datensaetze pruefen
            for ( recordNum = 0; recordNum < recordCount; ++recordNum )
            {
                // Spalten durchlaufen
                for ( fieldNum = 0; fieldNum < m_entrance2Table[num].count; ++fieldNum )
                {
                    switch( fieldNum )
                    {
                    case 0: // Day
                        if ( !CheckEntrance2Date( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos] ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;

                    case 1: // RefGroup
                    case 2: // RefTime
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos], 1, 9999 ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                    }

                    recordPos += m_entrance2TableField[num][fieldNum].size;
                }
            }
        }
        break;

    case 5:
        // Anzahl moeglicher Datensaetze
        if ( (recordCount < 1) /*|| (recordCount > (FLASH_SIZE_LIST / ENTRANCE_EXT_RECORD_SIZE_5))*/ )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Berechnete Listengroesze pruefen
        if ( size != (unsigned int)(recordCount * ENTRANCE2_RECORD_SIZE_5) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Ueberpruefen der Listendaten durchfuehren?
        if ( checkRecords )
        {
            // Datensaetze pruefen
            for ( recordNum = 0; recordNum < recordCount; ++recordNum )
            {
                // Spalten durchlaufen
                for ( fieldNum = 0; fieldNum < m_entrance2Table[num].count; ++fieldNum )
                {
                    switch ( fieldNum )
                    {
                    case 0: // RefReader
                    case 2: // RefAction
                    case 3: // RefTime
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos], 1, 9999 ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;

                    case 1: // PortIn
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos], 1, 3 ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                    }

                    recordPos += m_entrance2TableField[num][fieldNum].size;
                }
            }
        }
        break;

    case 6:
        // Anzahl moeglicher Datensaetze
        if ( (recordCount < 1) /*|| (recordCount > (FLASH_SIZE_LIST / ENTRANCE_EXT_RECORD_SIZE_6))*/ )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Berechnete Listengroesze pruefen
        if ( size != (unsigned int)(recordCount * ENTRANCE2_RECORD_SIZE_6) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Ueberpruefen der Listendaten durchfuehren?
        if ( checkRecords )
        {
            // Datensaetze pruefen
            for ( recordNum = 0; recordNum < recordCount; ++recordNum )
            {
                // Spalten durchlaufen
                for ( fieldNum = 0; fieldNum < m_entrance2Table[num].count; ++fieldNum )
                {
                    switch ( fieldNum )
                    {
                    case 0: // ID
                    case 1: // RefReader
                    case 4: // RefTime
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos], 1, 9999 ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;

                    case 2: // PortOut
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos], 1, 2 ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;

                    case 3: // Elapse
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_0 + recordPos], 0, 248 ) )
                        {
#ifdef _MSC_VER
                            //							TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                    }

                    recordPos += m_entrance2TableField[num][fieldNum].size;
                }
            }
        }
        break;

    case 7:
        // Anzahl moeglicher Datensaetze
        if ( (recordCount < 1) /*|| (recordCount > (FLASH_SIZE_LIST / ENTRANCE_EXT_RECORD_SIZE_7))*/ )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Berechnete Listengroesze pruefen
        if ( size != (unsigned int)(recordCount * ENTRANCE2_RECORD_SIZE_7) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Ueberpruefen der Listendaten durchfuehren?
        if ( checkRecords )
        {
            // Datensaetze pruefen
            for ( recordNum = 0; recordNum < recordCount; ++recordNum )
            {
                // Spalten durchlaufen
                for ( fieldNum = 0; fieldNum < m_entrance2Table[num].count; ++fieldNum )
                {
                    switch ( fieldNum )
                    {
                    case 0: // ID
                    case 1: // RefGroup
                    case 2: // RefTime
                    case 3: // RefReaderRelais
                    case 6: // RefReaderLED
                    case 7: // RefTimeRelais
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_7 + recordPos], 1, 9999 ) )
                        {
#ifdef _MSC_VER
                            //            TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;

                    case 4: // PortOut
                        if ( !CheckEntrance2Dec( (char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_7 + recordPos], 1, 2 ) )
                        {
#ifdef _MSC_VER
                            //            TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
#endif
                        }
                        break;

                    case 5: // Elapse
                        //            if (!CheckEntrance2Dec((char*)&buf[recordNum * ENTRANCE2_RECORD_SIZE_7 + recordPos], 0, 248)) {
                        //#ifdef _MSC_VER
                        //              TRACE("%s(%i):WARNING(1): CAEBuffer::CheckEntrance2List [  ]\n", __FILE__, __LINE__);
                        //#endif
                        //            }
                        break;
                    }

                    recordPos += m_entrance2TableField[num][fieldNum].size;
                }
            }
        }
        break;

    default:
        return -IV_ERROR_FUNCTION_PARAMETER;
    }

    return IV_ERROR_SUCCESS;
}

int Channel4Buffer::ImportEntrance2List( unsigned char num, unsigned short recordCount, const unsigned char *buf, unsigned int size )
{
    unsigned int i;

    // ListNum ist:
    // 0: Reader         (33 Byte)
    // 1: Identification (64 Byte)
    // 2: Location       (20 Byte)
    // 3: Time           (25 Byte)
    // 4: Holiday        (21 Byte)
    // 5: Events         (17 Byte)
    // 6: Action         (21 Byte)
    // 7: Action2        (39 Byte)

    // Zuvor reservierten Speicher freigeben
    if ( m_pucEntrance2ListMem[num] )
    {
        delete[] m_pucEntrance2ListMem[num];
        m_pucEntrance2ListMem[num] = NULL;
    }

    // Speicher reservieren
    m_pucEntrance2ListMem[num] = new unsigned char[size + sizeof(unsigned int)];

    if ( m_pucEntrance2ListMem[num] == NULL )
        return -IV_ERROR_LIST_PROCESSING;

    // Rohe Listendaten kopieren
    for ( i = 0; i < size; ++i )
        m_pucEntrance2ListMem[num][sizeof(unsigned int) + i] = buf[i]; // Platz fuer Anzahl Zeilen freihalten

    // Anzahl Listenzeilen eintragen
    *(unsigned int*)&m_pucEntrance2ListMem[num][0] = recordCount;
    //ChangeOrderLong(&m_pucEntranceListMem[ListNum][0]);

    // Listenspeichergroesse setzen
    m_ulEntrance2ListSize[num] = sizeof(unsigned int) + size;

    return IV_ERROR_SUCCESS;
}

unsigned int Channel4Buffer::ChangeOrderULong( unsigned int val )
{
    unsigned char buf[4];

    *(unsigned int *)buf = val;
    ChangeOrderLong( buf );
    return *(unsigned int *)buf;
} 

int Channel4Buffer::GetTableCount( unsigned char ucList )
{
    if ( ucList )
    {
        return MaxListB;
    }
    else
    {
        return MaxDatB;
    }
}

int Channel4Buffer::GetTableRecord( unsigned char ucNum, std::string& szName, 
                                   unsigned char *pucFieldCount, unsigned char *pucVariable, unsigned char ucList )
{

    if ( ucList )
    {
        // Bei Listen kann es nötig sein auf die Zutrittslistendaten zu verzweigen.
        if ( ucNum >= 100 )
        {
            // Auf Zutrtittskontrolllisten verzweigen
            if ( ucNum >= 100 && ucNum < 104 )
            {
                // Version 1 wird nicht durch diese Version unterstützt.
                return 0;
            }
            else if ( ucNum >= 200 && ucNum < 208 )
            {
                // Version 2
                *pucVariable = 0;
                return GetTableRecordEntrance2( ucNum, szName, pucFieldCount );
            }
        }

        if ( ucNum >= MaxListB )
            return 0;

        szName = ListenBeschreibung[ucNum].Name;
        *pucFieldCount = ListenBeschreibung[ucNum].Felder;
        *pucVariable = 0;
    }
    else
    {
        if ( ucNum >= MaxDatB )
            return 0;

        szName = DatensatzBeschreibung[ucNum].Name;
        *pucFieldCount = DatensatzBeschreibung[ucNum].Felder;
    }

    return 1;
}

int Channel4Buffer::GetTableField( unsigned char ucTableNum, unsigned char ucFieldNum, 
                                  std::string& szName, unsigned char *pucType, 
                                  unsigned char *pucLength, unsigned char ucList )
{

    if ( ucList )
    {
        // Bei Listen kann es nötig sein auf die Zutrittslistendaten zu verzweigen.
        if ( ucTableNum >= 100 )
        {
            // Auf Zutrittskontrolllisten verzweigen
            if ( ucTableNum >= 100 && ucTableNum < 104 )
            {
                // Version 1 wird durch diese Version nicht unterstützt.
                return 0;
            }
            else if ( ucTableNum >= 200 && ucTableNum < 208 )
            {
                // Version 2
                return GetTableFieldEntrance2( ucTableNum, ucFieldNum, szName, pucType, pucLength );
            }
        }

        if ( ucTableNum >= MaxListB )
            return 0;

        if ( ListenBeschreibung[ucTableNum].Felder <= ucFieldNum )
            return 0;

        szName = ListenFeld[ucTableNum][ucFieldNum].Name;
        *pucType   = ListenFeld[ucTableNum][ucFieldNum].Typ;
        *pucLength = ListenFeld[ucTableNum][ucFieldNum].Cnt;
    }
    else
    {
        if ( ucTableNum >= MaxDatB )
            return 0;

        if ( DatensatzBeschreibung[ucTableNum].Felder <= ucFieldNum )
            return 0;

        szName = DatensatzFeld[ucTableNum][ucFieldNum].Name;
        *pucType   = DatensatzFeld[ucTableNum][ucFieldNum].Typ;
        *pucLength = DatensatzFeld[ucTableNum][ucFieldNum].Cnt;
    }

    return 1;
}

int Channel4Buffer::GetTableRecordEntrance2( unsigned char ucNum, std::string& szName, 
                                            unsigned char *pucFieldCount )
{
    // Gueltigkeitsbereich pruefen
    if ( !(ucNum >= 200 && ucNum < 208) )
        return 0;

    // Erkennungsoffset abziehen
    ucNum -= 200;

    // Werte uebernehmen
    szName = m_entrance2Table[ucNum].name;
    *pucFieldCount = m_entrance2Table[ucNum].count;

    return 1;
}

int Channel4Buffer::GetTableFieldEntrance2( unsigned char ucTableNum, unsigned char ucFieldNum, 
                                           std::string& szName, unsigned char *pucType, 
                                           unsigned char *pucLength )
{
    // Gueltigkeitsbereich pruefen
    if ( !(ucTableNum >= 200 && ucTableNum < 208) )
        return 0;

    // Erkennungsoffset abziehen
    ucTableNum -= 200;
    /*
    // Gueltigkeitsbereich pruefen
    if (!((ucFieldNum >= 0) && (ucFieldNum < m_entrance2Table[ucTableNum].count))) {

    return 0;
    }
    */	
    // Werte uebernehmen
    szName = m_entrance2TableField[ucTableNum][ucFieldNum].name;
    *pucType = m_entrance2TableField[ucTableNum][ucFieldNum].type;
    *pucLength = m_entrance2TableField[ucTableNum][ucFieldNum].size;

    return 1;
}

void Channel4Buffer::DefaultIOModule( StructIOModule *pIOModule )
{
    int i;
    pIOModule->ucActiveU = 0x00;
    pIOModule->ucOnScreenU = 0x00;
    pIOModule->ucActiveI = 0x00;
    pIOModule->ucOnScreenI = 0x00;

    // Setzen der default Werte.
    for ( i = 0; i < 8; ++i )
    {
        // Spannungskanaele
        sprintf( (char*)pIOModule->szNameU[i], "U%i", i + 1 );
        sprintf( (char*)pIOModule->szUnitU[i], "V" );
        pIOModule->fMinU[i] = 0;
        pIOModule->fMaxU[i] = 10;

        // Stomkanaele
        sprintf( (char*)pIOModule->szNameI[i], "I%i", i + 1 );
        sprintf( (char*)pIOModule->szUnitI[i], "mA" );
        pIOModule->fMinI[i] = 0;
        pIOModule->fMaxI[i] = 20;
    }
}

// Groesse der Setupdaten im Flash ermitteln
unsigned int Channel4Buffer::GetFlashSetupSize()
{
    // ##############################################################
    // Pointer sind direkte Adressen die im AEIII+:
    //   1. Im Motorolla - Format vorliegen.
    //   2. Mit einem Basisoffset -FLASH_KONFIG_START- berechnet sind.

    // Pointer auf ListenBeschreibungen ermitteln
    unsigned int pLB = m_file.endian_reverse( m_file.m_flash_header.LIST_STRUCT );
    pLB -= FLASH_KONFIG_START;

    // Pointerarray der Listenbeschreibungen durchschreiten
    int i;
    for ( i = 0; i < MAX_DATENSATZ_BESCHREIBUNG; ++i )
    {
        // Nullpointer suchen fuer Listenanzahl
        if ( *(unsigned int*)&m_file.m_flash[pLB + i * sizeof(unsigned int)] == 0ul )
            break;
    }

    // Liste gefunden?
    if ( i > 0 )
    {
        // Pointer auf letzte definierte Liste
        pLB = *(unsigned int*)&m_file.m_flash[pLB + (i - 1) * sizeof(unsigned int)];

        //if (App->ByteOrder == BO_MOT) {

        ChangeOrderLong( (unsigned char*)&pLB );
        //}
        pLB -= FLASH_KONFIG_START;

        StructDatensatzBeschreibung *pLBStruct = (StructDatensatzBeschreibung*)&m_file.m_flash[pLB];

        // Pointer auf Feldarray der Liste
        unsigned int pFeldArray = pLBStruct->P;
        //if (App->ByteOrder == BO_MOT) {

        ChangeOrderLong( (unsigned char*)&pFeldArray );
        //}
        pFeldArray -= FLASH_KONFIG_START;

        // Felder vorhanden?
        if ( pLBStruct->Felder )
        {
            // Pointer auf letztes Feld der Listenbeschreibung ermitteln.
            unsigned int pFeld = *(unsigned int*)&m_file.m_flash[pFeldArray + (pLBStruct->Felder - 1) * sizeof(unsigned int)];
            //if (App->ByteOrder == BO_MOT) {

            ChangeOrderLong( (unsigned char*)&pFeld );
            //}
            pFeld -= FLASH_KONFIG_START;

            // Feldadresse + eigene Groesse bildet FlashSetupSize
            return pFeld + sizeof(StructDatensatzFeld);
        }


        // Keine Felder definiert!
        return pFeldArray;
    }

    // Keine Listenbeschreibung definiert!
    // Die Anzahl fuer Listenbeschreibungspointer muss mitberechnet werden,
    // um bei einer Uebertragung diese zu loeschen.
    return pLB + MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int);
}

unsigned int Channel4Buffer::GetMDEEKFlashSize_0400()
{
    // Erweiterten Struktur lesen
    StructExtendedFlash extBegin = *(StructExtendedFlash*)m_file.m_flashExtended;

    // Erweiterte Struktur vorhanden?
    if ( extBegin.check != EXTENDED_CHECK )
        return 0;

    // Strukturen und Eingabeketten fuer MDE lesen.
    if ( extBegin.offsetEventChains == 0 )
    {
        // Kein Eintrag bei neuem Dokument. Erst nach writeFlashImage
        // 1Byte Anzahl Eingabeketten fuer MDE
        return 1;
    }

    int Unterpunkte;
    unsigned int pEK;

    Unterpunkte = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetEventChains];
    //if (App->ByteOrder == BO_MOT) {

    ChangeOrderLong( (unsigned char*)&Unterpunkte );
    //}
    if ( Unterpunkte == 0 )
    {
        // 1Byte Anzahl Eingabeketten fuer MDE
        return 1;
    }

    // Pointer der letzten Eingabekette ermitteln
    pEK = extBegin.offsetEventChains + sizeof(unsigned char) + (Unterpunkte - 1) * sizeof(unsigned int);	
    pEK = *(unsigned int*)&m_file.m_flashExtended[pEK];
    //if (App->ByteOrder == BO_MOT) {

    ChangeOrderLong( (unsigned char*)&pEK );
    //}
    // 16kByte fuer FlashOffset + Basisadresse fuer Konfigdaten subtrahieren
    pEK -= (FLASH_SIZE_SU + FLASH_KONFIG_START);

    // vor jeder Eingabekettenstruktur liegt eine Bedien - Struktur
    pEK += sizeof(StructBedien);
    StructDatensatzEinlesen *pDE = (StructDatensatzEinlesen*)&m_file.m_flashExtended[extBegin.offsetEventChains + pEK];

    // Pointer auf Feldarray der Eingabekette
    unsigned int pFeldArray = pDE->P;
    //if (App->ByteOrder == BO_MOT) {

    ChangeOrderLong( (unsigned char*)&pFeldArray );
    //}
    pFeldArray -= (FLASH_SIZE_SU + FLASH_KONFIG_START);

    // Felder vorhanden?
    if ( pDE->Felder > 0 )
    {
        // Pointer auf letztes Feld der Eingabekette ermitteln.
        unsigned int pFeld = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetEventChains + pFeldArray + (pDE->Felder - 1) * sizeof(unsigned int)];
        //if (App->ByteOrder == BO_MOT) {

        ChangeOrderLong( (unsigned char*)&pFeld );
        //}
        pFeld -= (FLASH_SIZE_SU + FLASH_KONFIG_START);

        //StructFeldEinlesen *pDF = (StructFeldEinlesen*)&m_file.m_flashExtended[extBegin.offsetEventChains + pFeld];

        // Feldadresse + eigene Groesse bildet FlashSetupSize
        return pFeld + sizeof(StructFeldEinlesen);
    }

    // Keine Eingabekettenfelder definiert!
    return pFeldArray;
}

unsigned int Channel4Buffer::GetMDEEKFlashSize()
{
    // Erweiterten Struktur lesen
    StructExtendedFlash extBegin = *(StructExtendedFlash*)m_file.m_flashExtended;

    // Erweiterte Struktur vorhanden?
    if ( extBegin.check != EXTENDED_CHECK )
        return 0;
 
    // Strukturen und Eingabeketten fuer MDE lesen.
    if ( extBegin.offsetEventChains == 0 )
    {
        // Kein Eintrag bei neuem Dokument. Erst nach writeFlashImage
        // 4Byte Anzahl Eingabeketten fuer MDE
        return 4;

    }

    int Unterpunkte;
    unsigned int pEK;

    Unterpunkte = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetEventChains];
    //if (App->ByteOrder == BO_MOT) {

    ChangeOrderLong( (unsigned char*)&Unterpunkte );
    //}
    if ( Unterpunkte == 0 )
    {
        // 4Byte Anzahl Eingabeketten fuer MDE
        return 4;
    }

    // Pointer der letzten Eingabekette ermitteln
    pEK = extBegin.offsetEventChains + sizeof(unsigned int) + (Unterpunkte - 1) * sizeof(unsigned int);	
    pEK = *(unsigned int*)&m_file.m_flashExtended[pEK];
    //if (App->ByteOrder == BO_MOT) {

    ChangeOrderLong( (unsigned char*)&pEK );
    //}
    // 16kByte fuer FlashOffset + Basisadresse fuer Konfigdaten subtrahieren
    pEK -= (FLASH_SIZE_SU + FLASH_KONFIG_START);

    // vor jeder Eingabekettenstruktur liegt eine Bedien - Struktur
    pEK += sizeof(StructBedien);
    StructDatensatzEinlesen *pDE = (StructDatensatzEinlesen*)&m_file.m_flashExtended[extBegin.offsetEventChains + pEK];

    // Pointer auf Feldarray der Eingabekette
    unsigned int pFeldArray = pDE->P;
    //if (App->ByteOrder == BO_MOT) {

    ChangeOrderLong( (unsigned char*)&pFeldArray );
    //}
    pFeldArray -= (FLASH_SIZE_SU + FLASH_KONFIG_START);

    // Felder vorhanden?
    if ( pDE->Felder > 0 )
    {
        // Pointer auf letztes Feld der Eingabekette ermitteln.
        unsigned int pFeld = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetEventChains + pFeldArray + (pDE->Felder - 1) * sizeof(unsigned int)];
        //if (App->ByteOrder == BO_MOT) {

        ChangeOrderLong( (unsigned char*)&pFeld );
        //}
        pFeld -= (FLASH_SIZE_SU + FLASH_KONFIG_START);

        //StructFeldEinlesen *pDF = (StructFeldEinlesen*)&m_file.m_flashExtended[extBegin.offsetEventChains + pFeld];

        // Feldadresse + eigene Groesse bildet FlashSetupSize
        return pFeld + sizeof(StructFeldEinlesen);
    }

    // Keine Eingabekettenfelder definiert!
    return pFeldArray;
}

unsigned int Channel4Buffer::GetEntranceEKFlashSize()
{
    // Erweiterten Struktur lesen
    StructExtendedFlash extBegin = *(StructExtendedFlash*)m_file.m_flashExtended;

    // Erweiterte Struktur vorhanden?
    if ( extBegin.check != EXTENDED_CHECK )
        return 0;

    // Strukturen und Eingabeketten fuer Zutritt lesen.
    if ( extBegin.offsetAccessControlEventChainsV1 == 0 )
    {
        // Kein Eintrag bei neuem Dokument. Erst nach writeFlashImage
        // 4Byte Anzahl Eingabeketten fuer Zutritt
        return 4;
    }

    int Unterpunkte;
    unsigned int pEK;

    Unterpunkte = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV1];
    //if (App->ByteOrder == BO_MOT) {

    ChangeOrderLong((unsigned char*)&Unterpunkte);
    //}
    if ( Unterpunkte == 0 )
    {
        // 4Byte Anzahl Eingabeketten fuer Zutritt
        return 4;
    }

    // Pointer der letzten Eingabekette ermitteln
    pEK = extBegin.offsetAccessControlEventChainsV1 + sizeof(unsigned int) + (Unterpunkte - 1) * sizeof(unsigned int);	
    pEK = *(unsigned int*)&m_file.m_flashExtended[pEK];
    //if (App->ByteOrder == BO_MOT) {

    ChangeOrderLong((unsigned char*)&pEK);
    //}
    // 32kByte fuer FlashOffset + Basisadresse fuer Konfigdaten subtrahieren
    pEK -= (FLASH_SIZE_SU + EXTENDED_SIZE_SU + FLASH_KONFIG_START);

    // vor jeder Eingabekettenstruktur liegt eine Bedien - Struktur
    pEK += sizeof(StructBedien);
    StructDatensatzEinlesen *pDE = (StructDatensatzEinlesen*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV1 + pEK];

    // Pointer auf Feldarray der Eingabekette
    unsigned int pFeldArray = pDE->P;
    //if (App->ByteOrder == BO_MOT) {

    ChangeOrderLong( (unsigned char*)&pFeldArray );
    //}
    pFeldArray -= (FLASH_SIZE_SU + EXTENDED_SIZE_SU + FLASH_KONFIG_START);

    // Felder vorhanden?
    if ( pDE->Felder > 0 )
    {
        // Pointer auf letztes Feld der Eingabekette ermitteln.
        unsigned int pFeld = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV1 + pFeldArray + (pDE->Felder - 1) * sizeof(unsigned int)];
        //if (App->ByteOrder == BO_MOT) {

        ChangeOrderLong( (unsigned char*)&pFeld );
        //}
        pFeld -= (FLASH_SIZE_SU + EXTENDED_SIZE_SU + FLASH_KONFIG_START);

        //StructFeldEinlesen *pDF = (StructFeldEinlesen*)&m_file.m_flashExtended[extBegin.offsetEventChains + pFeld];

        // Feldadresse + eigene Groesse bildet FlashSetupSize
        return pFeld + sizeof(StructFeldEinlesen);
    }

    // Keine Eingabekettenfelder definiert!
    return pFeldArray;
}

unsigned int Channel4Buffer::GetEntrance2EKFlashSize()
{
    // Erweiterten Struktur lesen
    StructExtendedFlash extBegin = *(StructExtendedFlash*)m_file.m_flashExtended;

    // Erweiterte Struktur vorhanden?
    if ( extBegin.check != EXTENDED_CHECK )
        return 0;

    // Strukturen und Eingabeketten fuer Zutritt lesen.
    if ( extBegin.offsetAccessControlEventChainsV2 == 0 )
    {
        // Kein Eintrag bei neuem Dokument. Erst nach writeFlashImage
        // 4Byte Anzahl Eingabeketten fuer Zutritt
        return 4;
    }

    int Unterpunkte;
    unsigned int pEK;

    Unterpunkte = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2];
    //if (theApp.ByteOrder == BO_MOT) {

    //		ChangeOrderLong((unsigned char*)&Unterpunkte);
    //}
    if ( Unterpunkte == 0 )
    {
        // 4Byte Anzahl Eingabeketten fuer Zutritt
        return 4;
    }

    // Pointer der letzten Eingabekette ermitteln
    pEK = extBegin.offsetAccessControlEventChainsV2 + sizeof(unsigned int) + (Unterpunkte - 1) * sizeof(unsigned int);	
    pEK = *(unsigned int*)&m_file.m_flashExtended[pEK];
    //if (theApp.ByteOrder == BO_MOT) {

    //		ChangeOrderLong((unsigned char*)&pEK);
    //}
    // Basisadresse fuer Konfigdaten subtrahieren
    pEK -= FLASH_KONFIG_START;

    // vor jeder Eingabekettenstruktur liegt eine Bedien - Struktur
    pEK += sizeof(StructBedien);
    StructDatensatzEinlesen *pDE = (StructDatensatzEinlesen*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + pEK];

    // Pointer auf Feldarray der Eingabekette
    unsigned int pFeldArray = pDE->P;
    //if (theApp.ByteOrder == BO_MOT) {

    ChangeOrderLong( (unsigned char*)&pFeldArray );
    //}
    pFeldArray -= FLASH_KONFIG_START;

    // Felder vorhanden?
    if ( pDE->Felder > 0 )
    {
        // Pointer auf letztes Feld der Eingabekette ermitteln.
        unsigned int pFeld = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + pFeldArray + (pDE->Felder - 1) * sizeof(unsigned int)];
        //if (theApp.ByteOrder == BO_MOT) {

        ChangeOrderLong( (unsigned char*)&pFeld );
        //}
        pFeld -= FLASH_KONFIG_START;

        //StructFeldEinlesen *pDF = (StructFeldEinlesen*)&ExtendedMem[extBegin.offsetEventChains + pFeld];

        // Feldadresse + eigene Groesse bildet FlashSetupSize
        return pFeld + sizeof(StructFeldEinlesen);
    }

    // Keine Eingabekettenfelder definiert!
    return pFeldArray;
}

void Channel4Buffer::PrepareEntrance2EK( unsigned int offs, unsigned char *buf )
{
    // Erweiterten Struktur lesen
    int num;
    int numFeld;
    int posPointer = 0;
    int posData = 0;
    int posFeldArray;
    StructExtendedFlash extBegin = *(StructExtendedFlash*)m_file.m_flashExtended;

    // Erweiterte Struktur vorhanden?
    if ( extBegin.check != EXTENDED_CHECK )
        return;

    // Sind Keine Angaben zur ZK vorhanden?
    if ( extBegin.offsetAccessControlEventChainsV2 == 0 )
    {
        // Keine Unterpunkte
        *reinterpret_cast<unsigned int *>(buf) = 0;
        return;
    }

    int Unterpunkte;
    unsigned int pEK;

    Unterpunkte = *reinterpret_cast<unsigned int *>(&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2]);
    if ( Unterpunkte == 0 )
    {
        // 4Byte Anzahl Eingabeketten fuer Zutritt
        return;
    }

    // Anzahl Eingabeketten eintragen
    *reinterpret_cast<unsigned int *>(buf) = Unterpunkte;
    posPointer = sizeof(unsigned int);
    posData    = sizeof(unsigned int) + Unterpunkte * sizeof(unsigned int);;

    // Alle Eingabeketten durchlaufen
    for ( num = 0; num < Unterpunkte; ++num )
    {
        // Pointer der Eingabekette ermitteln
        pEK = extBegin.offsetAccessControlEventChainsV2 + sizeof(unsigned int) + num * sizeof(unsigned int);	
        pEK = *reinterpret_cast<unsigned int *>(&m_file.m_flashExtended[pEK]);
        // Basisadresse fuer Konfigdaten subtrahieren
        pEK -= FLASH_KONFIG_START;

        // Zeiger auf Eingabekette uebernehmen (ohne Basisoffset: FLASH_KONFIG_START)
        *reinterpret_cast<unsigned int *>(&buf[posPointer]) = pEK;
        posPointer += sizeof(unsigned int);


        // vor jeder Eingabekettenstruktur liegt eine Bedien - Struktur
        // diese uebernehmen
        memcpy( &buf[posData], &m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + pEK], sizeof(StructBedien) );
        posData += sizeof(StructBedien);
        // dann uebergehen
        pEK += sizeof(StructBedien);

        // Zeiger auf Struktur 
        StructDatensatzEinlesen *pDE = reinterpret_cast<StructDatensatzEinlesen *>(&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + pEK]);

        // Pointer auf Feldarray der Eingabekette
        unsigned int pFeldArray = pDE->P;
        ChangeOrderLong( reinterpret_cast<unsigned char *>(&pFeldArray) );

        pFeldArray -= FLASH_KONFIG_START;

        // Zeiger auf Feldarray anpassen und Struktur uebernehmen
        ChangeOrderLong( reinterpret_cast<unsigned char *>(&pDE->P) );
        pDE->P += offs;
        ChangeOrderLong( reinterpret_cast<unsigned char *>(&pDE->P) );
        memcpy(&buf[posData], pDE, sizeof(StructDatensatzEinlesen));
        ChangeOrderLong( reinterpret_cast<unsigned char *>(&pDE->P) );
        pDE->P -= offs;
        ChangeOrderLong( reinterpret_cast<unsigned char *>(&pDE->P) );
        posData += sizeof(StructDatensatzEinlesen);

        posFeldArray = posData;
        posData += pDE->Felder * sizeof(unsigned int);
        // Felder abarbeiten
        for ( numFeld = 0; numFeld < pDE->Felder; ++numFeld )
        {
            // Pointer auf Feld der Eingabekette ermitteln.
            unsigned int pFeld = *reinterpret_cast<unsigned int *>(&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + pFeldArray + numFeld * sizeof(unsigned int)]);
            ChangeOrderLong( reinterpret_cast<unsigned char *>(&pFeld) );

            // Zeiger auf Feldarray anpassen und Struktur uebernehmen
            *reinterpret_cast<unsigned int *>(&buf[posFeldArray]) = pFeld + offs;
            ChangeOrderLong( &buf[posFeldArray] );
            posFeldArray += sizeof(unsigned int);

            pFeld -= FLASH_KONFIG_START;

            StructFeldEinlesen *pDF = reinterpret_cast<StructFeldEinlesen *>(&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + pFeld]);

            // Feldstruktur uebernehmen
            memcpy( &buf[posData], pDF, sizeof(StructFeldEinlesen) );
            // Da die Strukturen der Felder hintereinander liegen ist zwischen den
            // Zeigern immer ein Unterschied von sizeof(StructFeldEinlesen)
            posData += sizeof(StructFeldEinlesen);
        }
    }
}

void Channel4Buffer::AdjustExtendedMemPointer()
{
    // Erweiterten Struktur lesen
    unsigned int num;
    int numFeld;
    StructExtendedFlash extBegin = *(StructExtendedFlash*)m_file.m_flashExtended;

    // Erweiterte Struktur vorhanden?
    if ( extBegin.check != EXTENDED_CHECK )
        return;

    unsigned int count;
    unsigned int rootOffset, childOffset, tmp, tmp2;
    ///////////////////////////////////////////////////////
    // MDE
    if ( extBegin.offsetEventChains != 0 )
    {
        // Zeiger der Eingabeketten fuer MDE anpassen
        count = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetEventChains];
        ChangeOrderLong( (unsigned char*)&count );

        // Jede Eingabekette anspringen
        for ( num = 0; num < count; ++num )
        {
            rootOffset = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetEventChains + (num + 1) * sizeof(unsigned int)];
            ChangeOrderLong( (unsigned char*)&rootOffset );

            tmp = rootOffset;

            if ( rootOffset > (FLASH_SIZE_SU + FLASH_KONFIG_START) )
            {
                rootOffset -= (FLASH_SIZE_SU + FLASH_KONFIG_START);
                tmp -= (FLASH_SIZE_SU + FLASH_KONFIG_START);
            }
            else
            {
                rootOffset += (FLASH_SIZE_SU + FLASH_KONFIG_START);
            }

            ChangeOrderLong( (unsigned char*)&rootOffset );
            *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetEventChains + (num + 1) * sizeof(unsigned int)] = rootOffset;

            StructBedien bedien = *(StructBedien*)&m_file.m_flashExtended[extBegin.offsetEventChains + tmp];
            ChangeOrderLong( (unsigned char*)&bedien.P );

            tmp2 = bedien.P;

            if ( bedien.P > (FLASH_SIZE_SU + FLASH_KONFIG_START) )
            {
                bedien.P -= (FLASH_SIZE_SU + FLASH_KONFIG_START);
                tmp2 -= (FLASH_SIZE_SU + FLASH_KONFIG_START);
            }
            else
            {
                bedien.P += (FLASH_SIZE_SU + FLASH_KONFIG_START);
            }

            ChangeOrderLong( (unsigned char*)&bedien.P );
            *(StructBedien*)&m_file.m_flashExtended[extBegin.offsetEventChains + tmp] = bedien;

            StructDatensatzEinlesen datensatzEinlesen = *(StructDatensatzEinlesen*)&m_file.m_flashExtended[extBegin.offsetEventChains + tmp2];
            ChangeOrderLong( (unsigned char*)&datensatzEinlesen.P );

            tmp = datensatzEinlesen.P;

            if ( datensatzEinlesen.P > (FLASH_SIZE_SU + FLASH_KONFIG_START) )
            {
                datensatzEinlesen.P -= (FLASH_SIZE_SU + FLASH_KONFIG_START);
                tmp -= (FLASH_SIZE_SU + FLASH_KONFIG_START);
            }
            else
            {
                datensatzEinlesen.P += (FLASH_SIZE_SU + FLASH_KONFIG_START);
            }

            ChangeOrderLong( (unsigned char*)&datensatzEinlesen.P );
            *(StructDatensatzEinlesen*)&m_file.m_flashExtended[extBegin.offsetEventChains + tmp2] = datensatzEinlesen;

            for ( numFeld = 0; numFeld < datensatzEinlesen.Felder; ++numFeld )
            {
                childOffset = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetEventChains + tmp + numFeld * sizeof(unsigned int)];
                ChangeOrderLong( (unsigned char*)&childOffset );

                if ( childOffset > (FLASH_SIZE_SU + FLASH_KONFIG_START) )
                {
                    childOffset -= (FLASH_SIZE_SU + FLASH_KONFIG_START);
                }
                else
                {
                    childOffset += (FLASH_SIZE_SU + FLASH_KONFIG_START);
                }

                ChangeOrderLong( (unsigned char*)&childOffset );
                *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetEventChains + tmp + numFeld * sizeof(unsigned int)] = childOffset;
            }
        }
    }

    ///////////////////////////////////////////////////////
    // Zutrittskontrolle Version 2

    /////////////////////////
    // Tabellen
    if ( extBegin.offsetAccessControlTablesV2 != 0 )
    {
        // Zeiger der Tabellen fuer ZK anpassen
        count = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetAccessControlTablesV2];

        // Jede Tabelle anspringen
        for ( num = 0; num < count; ++num )
        {
            rootOffset = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetAccessControlTablesV2 + (num + 1) * sizeof(unsigned int)];
            // Die Zeiger auf die einzelnen Tabellen sind 
            tmp = rootOffset;

            if ( rootOffset > FLASH_KONFIG_START )
            {
                rootOffset -= FLASH_KONFIG_START;
                tmp -= FLASH_KONFIG_START;
            }
            else
            {
                rootOffset += FLASH_KONFIG_START;
            }

            *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetAccessControlTablesV2 + (num + 1) * sizeof(unsigned int)] = rootOffset;

            StructDatensatzBeschreibung datensatzBeschreibung = *(StructDatensatzBeschreibung*)&m_file.m_flashExtended[extBegin.offsetAccessControlTablesV2 + tmp];
            ChangeOrderLong( (unsigned char*)&datensatzBeschreibung.P );

            tmp2 = datensatzBeschreibung.P;

            if ( datensatzBeschreibung.P > FLASH_KONFIG_START )
            {
                datensatzBeschreibung.P -= FLASH_KONFIG_START;
                tmp2 -= FLASH_KONFIG_START;
            }
            else
            {
                datensatzBeschreibung.P += FLASH_KONFIG_START;
            }

            ChangeOrderLong( (unsigned char*)&datensatzBeschreibung.P );
            *(StructDatensatzBeschreibung*)&m_file.m_flashExtended[extBegin.offsetAccessControlTablesV2 + tmp] = datensatzBeschreibung;

            // Zeiger auf die Felder abarbeiten
            for ( numFeld = 0; numFeld < datensatzBeschreibung.Felder; ++numFeld )
            {
                childOffset = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetAccessControlTablesV2 + tmp2 + numFeld * sizeof(unsigned int)];
                ChangeOrderLong( (unsigned char*)&childOffset );

                if ( childOffset > FLASH_KONFIG_START )
                {
                    childOffset -= FLASH_KONFIG_START;
                }
                else
                {
                    childOffset += FLASH_KONFIG_START;
                }

                ChangeOrderLong( (unsigned char*)&childOffset );
                *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetAccessControlTablesV2 + tmp2 + numFeld * sizeof(unsigned int)] = childOffset;
            }
        }
    }

    /////////////////////////
    // Eingabeketten
    if ( extBegin.offsetAccessControlEventChainsV2 != 0 )
    {
        // Zeiger der Eingabeketten fuer MDE anpassen
        count = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2];

        // Jede Eingabekette anspringen
        for ( num = 0; num < count; ++num )
        {
            rootOffset = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + (num + 1) * sizeof(unsigned int)];

            tmp = rootOffset;

            if ( rootOffset > FLASH_KONFIG_START )
            {
                rootOffset -= FLASH_KONFIG_START;
                tmp -= FLASH_KONFIG_START;
            }
            else
            {
                rootOffset += FLASH_KONFIG_START;
            }

            *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + (num + 1) * sizeof(unsigned int)] = rootOffset;

            StructBedien bedien = *(StructBedien*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + tmp];
            ChangeOrderLong( (unsigned char*)&bedien.P );

            tmp2 = bedien.P;

            if ( bedien.P > FLASH_KONFIG_START )
            {
                bedien.P -= FLASH_KONFIG_START;
                tmp2 -= FLASH_KONFIG_START;
            }
            else
            {
                bedien.P += FLASH_KONFIG_START;
            }

            ChangeOrderLong( (unsigned char*)&bedien.P );
            *(StructBedien*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + tmp] = bedien;

            StructDatensatzEinlesen datensatzEinlesen = *(StructDatensatzEinlesen*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + tmp2];
            ChangeOrderLong( (unsigned char*)&datensatzEinlesen.P );

            tmp = datensatzEinlesen.P;

            if ( datensatzEinlesen.P > FLASH_KONFIG_START )
            {
                datensatzEinlesen.P -= FLASH_KONFIG_START;
                tmp -= FLASH_KONFIG_START;
            }
            else
            {
                datensatzEinlesen.P += FLASH_KONFIG_START;
            }

            ChangeOrderLong( (unsigned char*)&datensatzEinlesen.P );
            *(StructDatensatzEinlesen*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + tmp2] = datensatzEinlesen;

            // Zeiger auf die Felder abarbeiten
            for ( numFeld = 0; numFeld < datensatzEinlesen.Felder; ++numFeld )
            {
                childOffset = *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + tmp + numFeld * sizeof(unsigned int)];
                ChangeOrderLong( (unsigned char*)&childOffset );

                if ( childOffset > FLASH_KONFIG_START )
                {
                    childOffset -= FLASH_KONFIG_START;
                }
                else
                {
                    childOffset += FLASH_KONFIG_START;
                }

                ChangeOrderLong( (unsigned char*)&childOffset );
                *(unsigned int*)&m_file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + tmp + numFeld * sizeof(unsigned int)] = childOffset;
            }
        }
    }
}

void Channel4Buffer::Entrance2Default()
{
    int table, field;

    // Pointerarray auf Strukturen initialisieren
    memset( m_entrance2Pointer, 0, sizeof(m_entrance2Pointer) );

    // List: Reader
    table = 0;
    m_entrance2Pointer[table] = MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int);

    m_entrance2Table[table].num = table;
    strncpy( m_entrance2Table[table].name, "Reader", sizeof( m_entrance2Table[table].name ) - 1 );
    m_entrance2Table[table].count = 6;
    m_entrance2Table[table].key = 0;
    m_entrance2Table[table].sort = 0;
    m_entrance2Table[table].flags = TF_SORT;

    // Id	Nummer	4	Int	Der primaere Schluessel der Reader - Tabelle. Er muss eindeutig sein.
    field = 0;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "Id", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // ZM	Nummer	4	Int	Gibt die zugehoerige Master ID an.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "ZM", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // TM	Nummer	3	Int	Gibt die Busnummer des Leser an.(intern 010..160, extern 011..161)
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "TM", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 4;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefLocation	Nummer	4	Int	Definiert die Raum ID, zu dem der ZL Zutritt verschafft.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefLocation", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefAction	Nummer	4	Int	Aktion, die bei gueltigem Zutritt ausgeloest wird.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefAction", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // PinGeneral	Nummer	8	Int	Mit der hier hinterlegten PIN erhaelt man ohne AW Zutritt.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "PinGeneral", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 9;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );


    // List: Identification
    table++;
    m_entrance2Pointer[table] = m_entrance2Pointer[table - 1] + sizeof(StructTable2) + ++field * sizeof(StructTable2Field);

    m_entrance2Table[table].num = table;
    strncpy( m_entrance2Table[table].name, "Identification", sizeof( m_entrance2Table[table].name ) - 1 );
    m_entrance2Table[table].count = 7;
    m_entrance2Table[table].key = 0;
    m_entrance2Table[table].sort = 0;
    m_entrance2Table[table].flags = TF_SORT;

    // Id	Text	20	ASCII	Gibt die zu pruefende Ausweisnummer an.
    field = 0;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "Id", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_A_NUM;
    m_entrance2TableField[table][field].size = 21;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // Group	Nummer	4	Int	Gibt die BG an, in welcher der Ausweis mit anderen enthalten ist.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "Group", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // Pin	Nummer	8	Int	Definiert die PIN. Falls 0 dann keine Abfrage
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "Pin", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 9;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // Menace	Nummer	4	Int	Definiert die Bedrohungsziffern (Erweiterungsziffern zum PIN). Falls 0 dann keine Abfrage
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "Menace", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // ActiveStart	Text	10	Date	Sperrbereich des AW (Ab wann ist der AW gueltig).
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "ActiveStart", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_A_NUM;
    m_entrance2TableField[table][field].size = 11;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // ActiveEnd	Text	10	Date	Sperrbereich des AW (Bis wann ist der AW gueltig).
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "ActiveEnd", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_A_NUM;
    m_entrance2TableField[table][field].size = 11;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // ActiveGeneral	Nummer	1	Int	Globale Sperrung des AW.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "ActiveGeneral", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 2;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );



    // List: Location
    table++;
    m_entrance2Pointer[table] = m_entrance2Pointer[table - 1] + sizeof(StructTable2) + ++field * sizeof(StructTable2Field);

    m_entrance2Table[table].num = table;
    strncpy( m_entrance2Table[table].name, "Location", sizeof( m_entrance2Table[table].name ) - 1 );
    m_entrance2Table[table].count = 4;
    m_entrance2Table[table].key = 0;
    m_entrance2Table[table].sort = 0;
    m_entrance2Table[table].flags = TF_SORT;

    // Id	Nummer	4	Int	Gibt den Raum an.
    field = 0;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "Id", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefGroup	Nummer	4	Int	Die BG fuer den Raum.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefGroup", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefTime	Nummer	4	Int	Das entsprechende Zeitmodell.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefTime", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefTimePin		Nummer	4	Int	Zeitmodell, wann eine PIN eingegeben werden muss.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefTimeNoPin", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );



    // List: Time
    table++;
    m_entrance2Pointer[table] = m_entrance2Pointer[table - 1] + sizeof(StructTable2) + ++field * sizeof(StructTable2Field);

    m_entrance2Table[table].num = table;
    strncpy( m_entrance2Table[table].name, "Time", sizeof( m_entrance2Table[table].name ) - 1 );
    m_entrance2Table[table].count = 4;
    m_entrance2Table[table].key = 0;
    m_entrance2Table[table].sort = 0;
    m_entrance2Table[table].flags = TF_SORT;

    // Id	Nummer	4	Int	ID des Zeitmodell.
    field = 0;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "Id", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // Weekdays	Nummer	7	Weekday	Definiert die Wochentage, an den das Zeitmodell gilt. (Format: Max. 7 Stellen 1 - 7 z.B. 134567= Montag, Mittwoch bis Sonntag)
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "Weekdays", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 8;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // TimeStart	Text	5	Time	Von wann gueltig. (Format 24h: HH:MM)
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "TimeStart", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_A_NUM;
    m_entrance2TableField[table][field].size = 6;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // TimeEnd	Text	5	Time	Bis wann gueltig.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "TimeEnd", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_A_NUM;
    m_entrance2TableField[table][field].size = 6;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );



    // List: Holiday
    table++;
    m_entrance2Pointer[table] = m_entrance2Pointer[table - 1] + sizeof(StructTable2) + ++field * sizeof(StructTable2Field);

    m_entrance2Table[table].num = table;
    strncpy( m_entrance2Table[table].name, "Holiday", sizeof( m_entrance2Table[table].name ) - 1 );
    m_entrance2Table[table].count = 3;
    m_entrance2Table[table].key = 0;
    m_entrance2Table[table].sort = 0;
    m_entrance2Table[table].flags = 0;

    // Day	Text	10	Date	Datum des Sperrtages. (Format: YYYY-MM-DD)
    field = 0;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "Day", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_A_NUM;
    m_entrance2TableField[table][field].size = 11;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefGroup	Nummer	4	Int	Gibt die BG fuer den Sperrtag an.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefGroup", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefTime	Nummer	4	Int	Gibt das abweichende Zeitmodell an.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefTime", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );



    // List: Event
    table++;
    m_entrance2Pointer[table] = m_entrance2Pointer[table - 1] + sizeof(StructTable2) + ++field * sizeof(StructTable2Field);

    m_entrance2Table[table].num = table;
    strncpy( m_entrance2Table[table].name, "Event", sizeof( m_entrance2Table[table].name ) - 1 );
    m_entrance2Table[table].count = 4;
    m_entrance2Table[table].key = 0;
    m_entrance2Table[table].sort = 0;
    m_entrance2Table[table].flags = 0;

    // RefReader	Nummer	4	Int	Master, an dem das TM ist.
    field = 0;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefReader", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // PortIn	Nummer	1	Int	Gibt den DI am TM an.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "PortIn", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 2;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefAction	Nummer	4	Int	ID der Aktion, die ausgefuehrt werden soll.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefAction", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefTime	Nummer	4	Int	Zeitmodell, wann der DI gueltig ist.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefTime", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );



    // List: Action
    table++;
    m_entrance2Pointer[table] = m_entrance2Pointer[table - 1] + sizeof(StructTable2) + ++field * sizeof(StructTable2Field);

    m_entrance2Table[table].num = table;
    strncpy( m_entrance2Table[table].name, "Action", sizeof( m_entrance2Table[table].name ) - 1 );
    m_entrance2Table[table].count = 5;
    m_entrance2Table[table].key = 0;
    m_entrance2Table[table].sort = 0;
    m_entrance2Table[table].flags = TF_SORT;

    // Id	Nummer	4	Int	ID der Aktion.
    field = 0;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "Id", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefReader	Nummer	4	Int	Master, an dem das Ereignis ausgeloest werden soll.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefReader", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset(  m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // PortOut	Nummer	1	Int	Welcher Ausgang wird geschaltet.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "PortOut", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 2;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // Elapse	Nummer	3	Int	Dauer des Schaltens in Sekunden, 0 -> dauernd
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "Elapse", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 4;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefTime	Nummer	4	Int	Zeitmodell, wann das Relais ausloesen soll.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefTime", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // List: Action2
    table++;
    m_entrance2Pointer[table] = m_entrance2Pointer[table - 1] + sizeof(StructTable2) + ++field * sizeof(StructTable2Field);

    m_entrance2Table[table].num = table;
    strncpy( m_entrance2Table[table].name, "Action2", sizeof( m_entrance2Table[table].name ) - 1 );
    m_entrance2Table[table].count = 8;
    m_entrance2Table[table].key = 0;
    m_entrance2Table[table].sort = 0;
    m_entrance2Table[table].flags = TF_SORT;

    // Id	Nummer	4	Int	ID der Aktion.
    field = 0;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "Id", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefGroup	Nummer	4	Int	Gibt die BG fuer die Relaisoeffnung an.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefGroup", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefTime	Nummer	4	Int	Zeitmodell, wann das Relais ausloesen soll.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefTime", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefReaderRelais	Nummer	4	Int	Master, an dem das Ereignis ausgeloest werden soll.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefReaderRelais", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // PortOut	Nummer	1	Int	Welcher Ausgang wird geschaltet.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "PortOut", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 2;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // Elapse	Nummer	6	Int	Dauer des Schaltens in Sekunden, 0 -> dauernd
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "Elapse", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_A_NUM;
    m_entrance2TableField[table][field].size = 7;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefReaderLED	Nummer	4	Int	Master, an dem das Ereignis ausgeloest werden soll.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefReaderLED", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

    // RefTimeRelais	Nummer	4	Int	Zeitmodell, wann das Relais ausloesen soll.
    field++;
    m_entrance2TableField[table][field].num = field;
    strncpy( m_entrance2TableField[table][field].name, "RefTimeRelais", sizeof( m_entrance2TableField[table][field].name ) - 1 );
    m_entrance2TableField[table][field].type = TF_NUM;
    m_entrance2TableField[table][field].size = 5;
    m_entrance2TableField[table][field].flags = 0;
    memset( m_entrance2TableField[table][field].format, 0, sizeof(m_entrance2TableField[table][field].format) );

}

void Channel4Buffer::ClearTimeboyListBuffer( unsigned char id, unsigned char num )
{
    unsigned char groupID, listNum;

    if ( id == 0xFF )
    {
        for ( groupID = 0; groupID < MAX_DATENSATZ_BESCHREIBUNG; ++groupID )
        {
            if ( num == 0xFF )
            {
                // Alle loeschen
                for ( listNum = 0; listNum < MAX_DATENSATZ_BESCHREIBUNG; ++listNum )
                {
                    m_ulTimeboyListSize[groupID][listNum] = 0;
                    m_timeboyFlags[groupID][listNum] = 0;

                    // Reservierten Speicher freigeben
                    if ( m_pucTimeboyListMem[groupID][listNum] )
                    {
                        delete[] m_pucTimeboyListMem[groupID][listNum];
                        m_pucTimeboyListMem[groupID][listNum] = NULL;
                    }
                }
            }
            else
            {
                if ( num >= MAX_DATENSATZ_BESCHREIBUNG )
                    return;

                m_ulTimeboyListSize[groupID][num] = 0;
                m_timeboyFlags[groupID][num] = 0;

                // Reservierten Speicher freigeben
                if ( m_pucTimeboyListMem[groupID][num] )
                {
                    delete[] m_pucTimeboyListMem[groupID][num];
                    m_pucTimeboyListMem[groupID][num] = NULL;
                }
            }
        }
    }
    else
    {
        if ( id >= MAX_DATENSATZ_BESCHREIBUNG )
            return;

        if ( num == 0xFF )
        {
            // Alle loeschen
            for ( listNum = 0; listNum < MAX_DATENSATZ_BESCHREIBUNG; ++listNum )
            {
                m_ulTimeboyListSize[id][listNum] = 0;
                m_timeboyFlags[id][listNum] = 0;

                // Reservierten Speicher freigeben
                if ( m_pucTimeboyListMem[id][listNum] )
                {
                    delete[] m_pucTimeboyListMem[id][listNum];
                    m_pucTimeboyListMem[id][listNum] = NULL;
                }
            }
        }
        else
        {
            if ( num >= MAX_DATENSATZ_BESCHREIBUNG )
                return;

            m_ulTimeboyListSize[id][num] = 0;
            m_timeboyFlags[id][num] = 0;

            // Reservierten Speicher freigeben
            if ( m_pucTimeboyListMem[id][num] )
            {
                delete[] m_pucTimeboyListMem[id][num];
                m_pucTimeboyListMem[id][num] = NULL;
            }
        }
    }
}

int Channel4Buffer::MakeTimeboyList( unsigned char id, unsigned char num, unsigned short recordCount, const unsigned char *buf, unsigned int size, int flags, int *error )
{
    unsigned int i;

    // Zuvor reservierten Speicher freigeben
    if ( m_pucTimeboyListMem[id][num] )
    {
        delete[] m_pucTimeboyListMem[id][num];
        m_pucTimeboyListMem[id][num] = NULL;
    }

    // Soll die Liste geloescht werden?
    if ( flags & 0x01 )
    {
        // Sicherstellen, dass keine unnoetigen Daten versendet werden.
        size = 0;
        recordCount = 0;
    }
    // Flags uebernehmen
    m_timeboyFlags[id][num] = flags;

    // Speicher reservieren
    m_pucTimeboyListMem[id][num] = new unsigned char[size + sizeof(unsigned int)];

    if ( m_pucTimeboyListMem[id][num] == NULL )
    {
        *error = IV_ERROR_LIST_PROCESSING;
        return 0;
    }

    // Rohe Listendaten kopieren
    for ( i = 0; i < size; ++i )
        m_pucTimeboyListMem[id][num][sizeof(unsigned int) + i] = buf[i]; // Platz fuer Anzahl Zeilen freihalten

    // Anzahl Listenzeilen eintragen
    *(unsigned int*)&m_pucTimeboyListMem[id][num][0] = recordCount;

    // Listenspeichergroesse setzen
    m_ulTimeboyListSize[id][num] = sizeof(unsigned int) + size;

    return 1;
}

bool Channel4Buffer::WriteListToBinFile( unsigned char listNum, const char *filePath )
{
    bool res = false;
    FILE *fp = NULL;

    if ( listNum > MAX_DATENSATZ_BESCHREIBUNG )
        return false;

    // Listendaten vorhanden?
    if ( m_listSize[listNum] == 0 )
        return false;

    unsigned char *pucBuf = &m_pucListMem[listNum][sizeof(unsigned int)];
    unsigned int size = m_listSize[listNum] - sizeof(unsigned int);
    //unsigned int recordCount = *(unsigned int*)m_pucListMem[listNum];

    // Daten Binaer rausschreiben
    fp = fopen( filePath, "wb" );
    if ( !fp )
        goto error_label_bin;

    fwrite( (const void*)pucBuf, 1, size, fp );

    // Schreiben in die Datei erfolgreich.
    res = true;

error_label_bin:
    if ( fp )
        fclose( fp );

    return res;
}

bool Channel4Buffer::WriteListToTextFile( unsigned char listNum, const char *filePath )
{
    bool res = false;
    std::string listName;
	unsigned char fieldCount, varList;
    std::string fieldName;
	unsigned char fieldNum, fieldType, fieldLength;
    int listOffs;
    FILE *fp = NULL;
    char line[512];

    // Listenbeschreibungen geladen?
    if ( GetTableCount( 1 ) == 0 )
        return false;

    // Listendaten vorhanden?
    if ( m_listSize[listNum] == 0 )
        return false;;

    unsigned char *pucBuf = &m_pucListMem[listNum][sizeof(unsigned int)];
    //unsigned int size = pChannel->m_listSize[listNum] - sizeof(unsigned int);
    unsigned int recordCount = *(unsigned int*)m_pucListMem[listNum];

    if ( !GetTableRecord( listNum, listName, &fieldCount, &varList, TRUE ) )
        goto error_label_txt;

    // Daten in Textform ablegen
    fp = fopen( filePath, "wb" );
    if ( !fp )
        goto error_label_txt;

    listOffs = 0;
    unsigned int recordNum;
    for ( recordNum = 0; recordNum < recordCount; ++recordNum )
    {
        memset( line, 0, sizeof(line) );
        for ( fieldNum = 0; fieldNum < fieldCount; ++fieldNum )
        {
            if ( !GetTableField( listNum, fieldNum, fieldName, &fieldType, &fieldLength, TRUE ) )
                goto error_label_txt;

            strcat( line, (char*)&pucBuf[listOffs] );
            if ( fieldNum != (fieldCount - 1) )
                strcat(line, "\t");

            listOffs += fieldLength;
        }

        fprintf( fp, "%s\x00D\x00A", line );
    }

    // Schreiben in Datei erfolgreich.
    res = true;

error_label_txt:
    if ( fp )
        fclose( fp );

    return res;
}

bool Channel4Buffer::WriteListEntrance2ToBinFile( unsigned char listNum, const char *filePath )
{
    bool res = false;
    FILE *fp = NULL;

    // Listendaten vorhanden?
    if ( m_ulEntrance2ListSize[listNum] == 0 )
        return false;

    unsigned char *pucBuf = &m_pucEntrance2ListMem[listNum][sizeof(unsigned int)];
    unsigned int size = m_ulEntrance2ListSize[listNum] - sizeof(unsigned int);
    //unsigned int recordCount = *(unsigned int*)m_pucEntrance2ListMem[listNum];

    // Daten Binaer rausschreiben
    fp = fopen( filePath, "wb" );
    if ( !fp )
        goto error_label_bin;

    fwrite( (const void*)pucBuf, 1, size, fp );

    // Daten erfolgreich geschrieben.
    res = true;

error_label_bin:
    if ( fp )
        fclose( fp );

    return res;
}

bool Channel4Buffer::WriteListEntrance2ToTextFile( unsigned char listNum, const char *filePath )
{
    bool res = false;
    std::string listName;
	unsigned char fieldCount, varList;
    std::string fieldName;
	unsigned char fieldNum, fieldType, fieldLength;
    int listOffs;
    FILE *fp = NULL;
    char line[512];

    // Listendaten vorhanden?
    if ( m_ulEntrance2ListSize[listNum] == 0 )
        return false;

    unsigned char *pucBuf = &m_pucEntrance2ListMem[listNum][sizeof(unsigned int)];
    //unsigned int size = m_ulEntrance2ListSize[listNum] - sizeof(unsigned int);
    unsigned int recordCount = *(unsigned int*)m_pucEntrance2ListMem[listNum];

    if ( !GetTableRecord( 200 + listNum, listName, &fieldCount, &varList, TRUE ) )
        goto error_label_txt;

    // Daten in Textform ablegen
    fp = fopen( filePath, "wb" );
    if ( !fp )
        goto error_label_txt;

    listOffs = 0;
    unsigned int recordNum;
    for ( recordNum = 0; recordNum < recordCount; ++recordNum )
    {
        memset( line, 0, sizeof(line) );
        for ( fieldNum = 0; fieldNum < fieldCount; ++fieldNum )
        {
            if ( !GetTableField( 200 + listNum, fieldNum, fieldName, &fieldType, &fieldLength, TRUE ) )
                goto error_label_txt;

            strcat( line, (char*)&pucBuf[listOffs] );
            if ( fieldNum != (fieldCount - 1) )
                strcat( line, "\t" );

            listOffs += fieldLength;
        }

        fprintf( fp, "%s\x00D\x00A", line );
    }

    // Daten erfolgreich geschrieben.
    res = true;

error_label_txt:
    if ( fp )
        fclose( fp );

    return res;
}

bool Channel4Buffer::WriteListTimeboyToBinFile( unsigned char listId, unsigned char listNum, const char *filePath )
{
    bool res = false;
    FILE *fp = NULL;

    // Listendaten vorhanden?
    if ( m_ulTimeboyListSize[listId][listNum] == 0 )
        return false;

    unsigned char *pucBuf = &m_pucTimeboyListMem[listId][listNum][sizeof(unsigned int)];
    unsigned int size = m_ulTimeboyListSize[listId][listNum] - sizeof(unsigned int);
    //unsigned int recordCount = *(unsigned int*)m_pucTimeboyListMem[listId][listNum];

    // Daten Binaer rausschreiben
    fp = fopen( filePath, "wb" );
    if ( !fp )
        goto error_label;

    fwrite( (const void*)pucBuf, 1, size, fp );
    fclose( fp );

    // Daten erfolgreich geschrieben.
    res = true;

error_label:
    if ( fp )
        fclose( fp );

    return res;
}

